/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Miroslav Kure <kurem@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gnome.h>
#include "data.h"
#include "functions.h"
#include "export.h"

static FILE *fw;
static FLAGS options;

/*
 *  Options
 */

void setup_options(void)
{
    options.info = 0;
    options.comment = 0;
    options.delatex = 1;
    options.picture = 0;
    options.size = 11;
}

void set_options_info(gboolean val) {
    options.info = (val)? 1 : 0;
}

void set_options_comment(gboolean val) {
    options.comment = (val)? 1 : 0;
}

void set_options_delatex(gboolean val) {
    options.delatex = (val)? 1 : 0;
}

void set_options_picture(gboolean val) {
    options.picture = (val)? 1 : 0;
}

void set_options_size(int val) {
    options.size = val;
}

gboolean get_options_info(void) {
    return options.info;
}

gboolean get_options_comment(void) {
    return options.comment;
}

gboolean get_options_delatex(void) {
    return options.delatex;
}

gboolean get_options_picture(void) {
    return options.picture;
}

int get_options_size(void) {
    return options.size;
}

/*
 *  Functions
 */

static gint bidir(Transition *trans)
{
    GSList *l;
    int i = 0;
    for (l = aut.transitions, i = 0; l != NULL; l = l->next, i++) {
	Transition *t = l->data;
	if (trans->from == t->to && trans->to == t->from)
	    return i;
    }
    return -1;
} /* bidir */

/*
 *  LaTeX
 */

static void write_doc_header(void)
{
    fprintf(fw, "\\documentclass[a4paper, %ipt]{article}\n", options.size);
    fprintf(fw, "\\usepackage{czech}\n\n");
    fprintf(fw, "\\begin{document}\n");
} /* write_doc_header */

static void write_doc_footer(void)
{
    fprintf(fw, "\\end{document}\n");
} /* write_doc_footer */

static void write_rules_table(void)
{
    GSList *l;
    gchar *text, *tmp;

    fprintf(fw, "\\begin{tabular}{c c l l c l}\n");

    text = get_gram_description(TRUE);
    fprintf(fw, "G & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_symbols(gram.nonterminal_symbols, TRUE);
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "N & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_symbols(gram.terminal_symbols, TRUE);
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "T & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    fprintf(fw, "P & = & \\multicolumn{4}{l}{\\{}\\\\\n");

    for (l = gram.rules; l != NULL; l = l->next) {
	Rule *rule = l->data;

	tmp = get_string_of_symbols(rule->left, FALSE);
	text = (options.delatex == 1)? delatex(tmp): g_strdup(tmp);
	g_free(tmp);
	fprintf(fw, " & & & %s & $\\to$ & ", text);
	g_free(text);

	tmp = get_string_of_symbols(rule->right, FALSE);
	text = (options.delatex == 1)? delatex(tmp): g_strdup(tmp);
	g_free(tmp);
	fprintf(fw, "%s \\\\\n", text);
	g_free(text);
    }
    fprintf(fw, " & & \\multicolumn{4}{l}{\\}}\\\\\n");
    fprintf(fw, "\\end{tabular}\n");
} /* write_rules_table */

static void write_grammar_table(void)
{
    write_doc_header();
    if (options.comment == 1) {
	fprintf(fw, get_comment());
	fprintf(fw, "\n\n");
    }

    write_rules_table();
    write_doc_footer();
} /* write_grammar_table */

static void write_stack_transitions_table(void)
{
    GSList *l;
    gchar *text, *tmp;

    fprintf(fw, "\\begin{tabular}{c c l l l l}\n");

    text = get_stack_aut_description(TRUE);
    fprintf(fw, "A & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_symbols(aut.entry_symbols, TRUE);
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "T & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_states();
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "Q & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_symbols(aut.stack_symbols, TRUE);
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "Z & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_ending_states();
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "F & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    fprintf(fw, "$\\delta$ & = & \\multicolumn{4}{l}{\\{} \\\\\n");

    for (l = aut.transitions; l != NULL; l = l->next) {
	Transition *trans = l->data;

	text = delatex(trans->from->name);
	fprintf(fw, " & & & $\\delta$ (%s, ", text);
	g_free(text);

	tmp = get_string_of_symbols(trans->entry_symbols, FALSE);
	text = (options.delatex == 1)? delatex(tmp): g_strdup(tmp);
	g_free(tmp);
	fprintf(fw, "%s, ", text);
	g_free(text);

	tmp = get_string_of_symbols(trans->from_stack, FALSE);
	text = (options.delatex == 1)? delatex(tmp): g_strdup(tmp);
	g_free(tmp);
	fprintf(fw, "%s) & = &", text);
	g_free(text);

	text = debrack(trans->to->name);
	fprintf(fw, " (%s,  ", text);
	g_free(text);

	tmp = get_string_of_symbols(trans->to_stack, FALSE);
	text = (options.delatex == 1)? delatex(tmp): g_strdup(tmp);
	g_free(tmp);
	fprintf(fw, "%s) \\\\\n", text);
	g_free(text);
    }
    fprintf(fw, " & & \\multicolumn{4}{l}{\\}}\\\\\n");
    fprintf(fw, "\\end{tabular}\n");
} /* write_stack_transitions_table */

static void write_finite_transitions_table(void)
{
    GSList *l;
    gchar *text, *tmp;

    fprintf(fw, "\\begin{tabular}{c c l l l l}\n");

    text = get_fin_aut_description(TRUE);
    fprintf(fw, "A & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_symbols(aut.entry_symbols, TRUE);
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "T & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_states();
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "Q & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    tmp = get_string_of_ending_states();
    text = debrack(tmp); g_free(tmp);
    fprintf(fw, "F & = & \\multicolumn{4}{l}{%s}\\\\\n", text);
    g_free(text);

    fprintf(fw, "$\\delta$ & = & \\multicolumn{4}{l}{\\{}\\\\\n");

    for (l = aut.transitions; l != NULL; l = l->next) {
	Transition *trans = l->data;

	text = debrack(trans->from->name);
	fprintf(fw, " & & & $\\delta$(%s,  ", text);
	g_free(text);

	tmp = get_string_of_symbols(trans->entry_symbols, FALSE);
	text = (options.delatex == 1)? delatex(tmp): g_strdup(tmp);
	g_free(tmp);
	fprintf(fw, "%s) & = & ", text);
	g_free(text);

	text = debrack(trans->to->name);
	fprintf(fw, " %s \\\\\n ", text);
	g_free(text);
    }
    fprintf(fw, " & & \\multicolumn{4}{l}{\\}}\\\\\n");
    fprintf(fw, "\\end{tabular}\n");
} /* write_finite_transitions_table */

static void write_automata_info(void)
{
    gchar *tmp, *text;

    fprintf(fw, "\\begin{tabular}{l l}\n");
    if (get_current() == STACK) {
	tmp = get_string_of_symbols(stack, FALSE);
	text = (options.delatex == 1)? delatex(tmp): g_strdup(tmp);
	g_free(tmp);
	fprintf(fw, "stack: & %s\\\\\n", text);
	g_free(text);
    }

    tmp = get_string_of_symbols(symbols, FALSE);
    text = (options.delatex == 1)? delatex(tmp): g_strdup(tmp);
    g_free(tmp);
    fprintf(fw, "sentence: & %s\\\\\n\\end{tabular}\n", text);
    g_free(text);
} /* write_automata_info */

static void write_automata_table(void)
{
    write_doc_header();

    if (options.comment == 1) {
	fprintf(fw, get_comment());
	fprintf(fw, "\n\n");
    }

    if (get_current() == STACK)
	write_stack_transitions_table();
    else
	write_finite_transitions_table();

    if (options.info == 1)
	write_automata_info();

    write_doc_footer();
} /* write_automata_table */

/*
 *  Metapost
 */

static void write_pic_header(void)
{
    fprintf(fw, "input boxes;\n\n");
    fprintf(fw, "vardef gaae_q (suffix s)(expr t, p) =\n  circleit.s (t);\n  s.dx = s.dy = 1.8mm;\n s.c = p;\nenddef;\n\n");
    fprintf(fw, "vardef gaae_draw_q (text all) =\n  forsuffixes s = all:\n    drawboxed (s);\n  endfor\nenddef;\n\n");
    fprintf(fw, "vardef gaae_draw_f (text all) =\n  forsuffixes s = all:\n    transform tt;\n    s.c transformed tt = s.c;\n    s.w transformed tt = s.w + (.5mm, 0);\n    s.n transformed tt = s.n + (0, -.5mm);\n    draw bpath.s transformed tt;\n    drawboxed (s);\n  endfor\nenddef;\n\n");
    fprintf(fw, "vardef gaae_sag (suffix a, b)(expr s) =\n  a.c ... (.5 [a.c, b.c] shifted s) ... b.c\nenddef;\n\n");
    fprintf(fw, "vardef gaae_draw_d (suffix a, b)(expr p)(suffix pos)(expr t) =\n  path pp;\n  pp := p cutbefore bpath a cutafter bpath b;\n  drawarrow pp;\n  label.pos (t, point .5 * length pp of pp);\nenddef;\n\n");
    fprintf(fw, "vardef gaae_draw_c (suffix s)(expr angle)(suffix pos)(expr t) =\n  transform tt;\n  path pp;\n  s.c transformed tt = s.c;\n  s.w transformed tt = s.w + (-1.75mm, 0);\n  s.n transformed tt = s.n + (0, 1.75mm);\n  pp := (fullcircle xscaled 5mm yscaled 5mm xscaled -1)\n      rotated angle shifted\n      ((bpath s transformed tt) intersectionpoint (s.c -- (s.c shifted ((20cm, 0) rotated angle))))\n      cutbefore bpath s cutafter bpath s;\n  drawarrow pp;\n  label.pos (t, point .5 * length pp of pp);\nenddef;\n\n");
    fprintf(fw, "beginfig (1)\n  u=0.25mm;\n");
} /* write_pic_header */

static void write_pic_footer(void)
{
    fprintf(fw, "endfig;\n\nbye\n");
} /* write_pic_footer */

static void write_pic_body(void)
{
    GSList *l;
    double x, y, fx, fy;
    int sig = '+';
    int index = 0;

    for (l = aut.states; l != NULL; l = l->next) {
	State *state = l->data;
	gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(state->group),
				     &x, &y, &fx, &fy);
	fprintf(fw,
		"  gaae_q (%s, btex $%s$ etex, (%f*u, %f*u));\n",
		state->name, state->name, x, 1200-y);
	if (state->atribute != NULL && strstr(state->atribute, "e"))
	    fprintf(fw, "  gaae_draw_f (%s);\n", state->name);
	else
	    fprintf(fw, "  gaae_draw_q (%s);\n", state->name);
    }

    for (l = aut.transitions, index = 0; l != NULL; l = l->next, index++) {
	Transition *trans = l->data;
	gchar *symbols = get_string_of_entry_symbols(trans);

	if (trans->from == trans->to)
	    fprintf(fw,
		    "  gaae_draw_c (%s, 180, lft, btex $%s$ etex);\n",
		    trans->from->name, symbols);
	else {
	    int i;
	    if ((i = bidir(trans)) != -1) {
		sig = (i < index)? '+': '-';

		fprintf(fw,
			"  gaae_draw_d (%s, %s, gaae_sag (%s, %s, (%c10u, %c10u)),",
			trans->from->name, trans->to->name,
			trans->from->name, trans->to->name, sig, sig);
		fprintf(fw, "lrt, btex $%s$ etex);\n", symbols);
	    } else {
		fprintf(fw,
			"  gaae_draw_d (%s, %s, gaae_sag (%s, %s, (0, 0)),",
			trans->from->name, trans->to->name,
			trans->from->name, trans->to->name);
		fprintf(fw, "lrt, btex $%s$ etex);\n", symbols);
	    }
	}
	g_free(symbols);
    }
} /* write_pic_body */

static void write_automata_picture(void)
{
    write_pic_header();
    write_pic_body();
    write_pic_footer();
} /* write_automata_picture */

gboolean export(char *fname)
{
    g_return_val_if_fail(fname != NULL, FALSE);

    if ((fw = fopen(fname, "w")) == NULL)
	return FALSE;

    switch (get_current()) {
    case REGULAR:
    case CONTEXTFREE:
    case GRAMMAR:
	write_grammar_table();
	break;
    case STACK:
	write_automata_table();
	break;
    case FINITE:
	if (options.picture == 0)
	    write_automata_table();
	else
	    write_automata_picture();
	break;
    default:
	return FALSE;
    }

    fclose(fw);
    return TRUE;
} /* export */

