/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Miroslav Kure <kurem@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gnome.h>
#include <stdlib.h>
#include "data.h"
#include "graphics.h"
#include "functions.h"

/*
 *    Data section - main structures
 */

automata aut;
grammar gram;

static KIND current;
static Transition *aux_trans;
static Rule *aux_rule;

GSList *states;
GSList *symbols;
GSList *rules;
GSList *stack;
GSList *pairs;

static GSList *coordinates;
static GString *comment;

/*
 *    Functions for manipulating data structures
 */

void setup_comment(void)
{
    comment = g_string_sized_new(20);
} /* setup_comment */

void set_comment(gchar *text)
{
    comment = g_string_truncate(comment, 0);
    comment = g_string_append(comment, text);
} /* set_comment */

gchar *get_comment(void)
{
    return comment->str;
} /* get_comment */

void set_current(KIND kind)
{
    current = kind;
} /* set_current */

KIND get_current(void)
{
    return current;
} /* get_current */

void delete_state(State *state)
{
    g_free(state->name);
    if (state->atribute != NULL)
	g_free(state->atribute);
    if (state->group != NULL)
	gtk_object_destroy(GTK_OBJECT(state->group));	
    
    g_free(state);
} /* delete_state */

void delete_transition(Transition *trans)
{
    if (trans->group != NULL)
	gtk_object_destroy(GTK_OBJECT(trans->group));
    
    if (trans->entry_symbols != NULL)
	g_slist_free(trans->entry_symbols);

    if (trans->from_stack != NULL)
	g_slist_free(trans->from_stack);

    if (trans->to_stack != NULL)
	g_slist_free(trans->to_stack);

    g_free(trans);
} /* delete_transition */

void delete_symbol(Symbol *symbol)
{
    g_free(symbol->name);
    if (symbol->atribute != NULL)
	g_free(symbol->atribute);
    g_free(symbol);
} /* delete_symbol */

void delete_rule(Rule *rule) /* CHECK: symboly!!! */
{
    g_slist_free(rule->left);
    g_slist_free(rule->right);
    g_free(rule);
} /* delete_rule */

State *add_state(int id, char *name, char *atribute)
{
    State *state;

    state = g_new(State, 1);

    state->id = id;
    state->name = g_strdup(name);
    if (atribute != NULL)
	state->atribute = g_strdup(atribute);
    else
	state->atribute = NULL;
    state->group = NULL;

    aut.states = g_slist_append(aut.states, state);

    return state;
} /* add_state */

gboolean add_symbol(int id, gchar *name, gchar *atribute, gchar *type)
{
    Symbol *symbol = g_new(Symbol, 1);
    symbol->id = id;
    symbol->name = g_strdup(name);
    symbol->atribute = g_strdup(atribute);

    if (strcmp(type, "entry_symbol") == 0) {
	aut.entry_symbols = g_slist_append(aut.entry_symbols, symbol);
	return TRUE;
    } else if (strcmp(type, "stack_symbol") == 0) {
	aut.stack_symbols = g_slist_append(aut.stack_symbols, symbol);
	return TRUE;
    } else if (strcmp(type, "terminal_symbol") == 0) {
	gram.terminal_symbols = g_slist_append(gram.terminal_symbols, symbol);
	return TRUE;
    } else if (strcmp(type, "nonterminal_symbol") == 0) {
	gram.nonterminal_symbols = g_slist_append(gram.nonterminal_symbols,
						  symbol);
	return TRUE;
    }
    return FALSE;
} /* add_symbol */

void add_state_to_list(int id)
{
    State *state;
    state = find_state_by_id(id);

    g_return_if_fail(state != NULL);

    states = g_slist_append(states, state);
} /* add_state_to_list */

void add_symbol_to_list(int id, gchar *type)
{
    Symbol *symbol = NULL;

    if (strcmp(type, "sentence") == 0)
	symbol = find_symbol_by_id(id, aut.entry_symbols);
    else if (strcmp(type, "terminal_symbol") == 0)
	symbol = find_symbol_by_id(id, gram.terminal_symbols);
    else if (strcmp(type, "nonterminal_symbol") == 0)
	symbol = find_symbol_by_id(id, gram.nonterminal_symbols);

    g_return_if_fail(symbol != NULL);
    
    symbols = g_slist_append(symbols, symbol);
} /* add_symbol_to_list */

void add_to_stack(int id)
{
    Symbol *symbol;
    symbol = find_symbol_by_id(id, aut.stack_symbols);

    g_return_if_fail(symbol != NULL);
    
    stack = g_slist_append(stack, symbol);
} /* add_to_stack */

void add_coordinate(int id, double x, double y)
{
    GSList *l;
    for (l = coordinates; l != NULL; l = l->next) {
	Coordinate *coordinate = l->data;
	if (coordinate->id == id) {
	    coordinate->x = x;
	    coordinate->y = y;
	    break;
	}
    }
    if (l == NULL) {
	Coordinate *coordinate = g_new(Coordinate, 1);
	coordinate->id = id;
	coordinate->x = x;
	coordinate->y = y;

	coordinates = g_slist_append(coordinates, coordinate);
    }
} /* add_coordinate */

void setup_aux(void)
{
    KIND type = get_current();
    if (type == FINITE || type == STACK) {
	aux_trans = g_new(Transition, 1);
	aux_trans->from = aux_trans->to = NULL;
	aux_trans->entry_symbols = NULL;
	aux_trans->from_stack = aux_trans->to_stack = NULL;
	aux_trans->group = NULL;
    } else if (type == REGULAR || type == CONTEXTFREE || type == GRAMMAR) {
	aux_rule = g_new(Rule, 1);
	aux_rule->left = NULL;
	aux_rule->right = NULL;
    }
} /* setup_aux */

void add_aux_to_list(gchar *extra)
{
    KIND type = get_current();
    GSList *l;

    if (type == FINITE) {
	for (l = aut.transitions; l != NULL; l = l->next) {
	    Transition *transition = l->data;
	    if (transition->from == aux_trans->from) {
		if (transition->to == aux_trans->to) {
		    transition->entry_symbols = g_slist_append(
			transition->entry_symbols,
			aux_trans->entry_symbols->data);
		    delete_transition(aux_trans);
		    return;
		}
	    }
	}
	aut.transitions = g_slist_append(aut.transitions, aux_trans);
	aux_trans = NULL;
    } else if (type == STACK) {
	if (extra == NULL) {
	    aut.transitions = g_slist_append(aut.transitions, aux_trans);
	    aux_trans = NULL;
	} else {
	    Pair *pair = g_new(Pair, 1);
	    pair->stack = stack;
	    if (states != NULL)
		pair->state = states->data;
	    else
		pair->state = NULL;
	    pairs = g_slist_append(pairs, pair);
	    states = NULL;
	    stack = NULL;
	}
    } else if (type == REGULAR || type == CONTEXTFREE || type == GRAMMAR) {
	if (extra == NULL)
	    gram.rules = g_slist_append(gram.rules, aux_rule);
	else
	    rules = g_slist_append(rules, aux_rule);
	aux_rule = NULL;
    }	
} /* add_aux_to_list */

void add_state_to_transition(int id)
{
    if (aux_trans->from == NULL) {
	aux_trans->from = find_state_by_id(id);
	g_return_if_fail(aux_trans->from != NULL);
    }
    else if (aux_trans->to == NULL) {
	aux_trans->to = find_state_by_id(id);
	g_return_if_fail(aux_trans->to != NULL);
    }
} /* add_state_to_transition */

void add_entry_symbol_to_transition(int id)
{
    if (id != 0) {
	Symbol *symbol = find_symbol_by_id(id, aut.entry_symbols);
	g_return_if_fail(symbol != NULL);
	
	aux_trans->entry_symbols = g_slist_append(aux_trans->entry_symbols,
						  symbol);
    }
}  /* add_entry_symbol_to_transition */

void add_stack_symbol_to_transition(int id, gchar *direction)
{
    Symbol *symbol = find_symbol_by_id(id, aut.stack_symbols);
    
    g_return_if_fail(symbol != NULL);

    if (strcmp(direction, "from_stack") == 0)
	aux_trans->from_stack = g_slist_append(aux_trans->from_stack, symbol);
    else if (strcmp(direction, "to_stack") == 0)
	aux_trans->to_stack = g_slist_append(aux_trans->to_stack, symbol);
} /* add_stack_symbol_to_transition */

void add_symbol_to_rule(int id, gchar *type, gchar *side)
{
    Symbol *symbol = NULL;
    if (strcmp(type, "terminal_symbol") == 0)
        symbol = find_symbol_by_id(id, gram.terminal_symbols);
    else if (strcmp(type, "nonterminal_symbol") == 0)
        symbol = find_symbol_by_id(id, gram.nonterminal_symbols);

    g_return_if_fail(symbol != NULL);

    if (strcmp(side, "right") == 0)
	aux_rule->right = g_slist_append(aux_rule->right, symbol);
    else if (strcmp(side, "left") == 0)
	aux_rule->left = g_slist_append(aux_rule->left, symbol);
} /* add_symbol_to_rule */

State *find_state_by_group(GnomeCanvasGroup *group)
{
    GSList *l;
    State *state;

    for (l = aut.states; l != NULL; l = l->next) {
	state = l->data;
	if (state->group == group)
	    return state;
    }
    return NULL;
} /* find_state_by_group */

State *find_state_by_id(int id)
{
    GSList *l;
    State *state;

    for (l = aut.states; l != NULL; l = l->next) {
	state = l->data;
	if (state->id == id)
	    return state;
    }
    return NULL;
} /* find_state_by_id */

State *find_state_by_name(gchar *name)
{
    GSList *l;
    State *state;

    for (l = aut.states; l != NULL; l = l->next) {
	state = l->data;
	if (strcmp(state->name, name) == 0)
	    return state;
    }
    return NULL;
} /* find_state_by_name */

Symbol *find_symbol_by_id(int id, GSList *list)
{
    GSList *l;
    Symbol *symbol;

    for (l = list; l != NULL; l = l->next) {
	symbol = l->data;
	if (symbol->id == id)
	    return symbol;
    }
    return NULL;
} /* find_symbol_by_id */

Symbol *find_symbol_by_name(gchar *name, GSList *list)
{
    GSList *l;
    Symbol *symbol;

    for (l = list; l != NULL; l = l->next) {
	symbol = l->data;
	if (strcmp(symbol->name, name) == 0)
	    return symbol;
    }
    return NULL;
} /* find_symbol_by_name */

Transition *find_transition_by_group(GnomeCanvasGroup *group)
{
    GSList *l;
    Transition *trans;

    for (l = aut.transitions; l != NULL; l = l->next) {
	trans = l->data;
	if (trans->group == group)
	    return trans;
    }
    return NULL;
} /* find_transition_by_group */

Coordinate *find_coordinate_by_id(int id)
{
    GSList *l;
    Coordinate *coordinate;

    for (l = coordinates; l != NULL; l = l->next) {
	coordinate = l->data;
	if (coordinate->id == id)
	    return coordinate;
    }

    return NULL;
} /* find_coordinate_by_id */

void clear_other_data(void)
{
    GSList *l;

    g_slist_free(states); states = NULL; /* CHECK: delete items */
    g_slist_free(symbols); symbols = NULL;
    g_slist_free(stack); stack = NULL;

    for (l = rules; l != NULL; l = l->next) {
	Rule *rule = l->data;
	delete_rule(rule);
    }
    g_slist_free(rules); rules = NULL;

    for (l = pairs; l != NULL; l = l->next) {
	Pair *pair = l->data;
	g_slist_free(pair->stack);
	g_free(pair);
    }
    g_slist_free(pairs); pairs = NULL;

/* and all remaining auxiliary lists... */
} /* clear_other_data */

void clear_all_data(void)
{
    GSList *l;
    for (l = aut.transitions; l != NULL; l = l->next) {
	Transition *trans = l->data;
	delete_transition(trans);
    }
    g_slist_free(aut.transitions); aut.transitions = NULL;
    
    for (l = aut.states; l != NULL; l = l->next) {
	State *state = l->data;
	delete_state(state);
    }
    g_slist_free(aut.states); aut.states = NULL;
    
    for (l = aut.entry_symbols; l != NULL; l = l->next) {
	Symbol *symbol = l->data;
	delete_symbol(symbol);
    }
    g_slist_free(aut.entry_symbols); aut.entry_symbols = NULL;

    for (l = aut.stack_symbols; l != NULL; l = l->next) {
	Symbol *symbol = l->data;
	delete_symbol(symbol);
    }
    g_slist_free(aut.stack_symbols); aut.stack_symbols = NULL;

    for (l = gram.nonterminal_symbols; l != NULL; l = l->next) {
	Symbol *symbol = l->data;
	delete_symbol(symbol);
    }
    g_slist_free(gram.nonterminal_symbols); gram.nonterminal_symbols = NULL;

    for (l = gram.terminal_symbols; l != NULL; l = l->next) {
	Symbol *symbol = l->data;
	delete_symbol(symbol);
    }
    g_slist_free(gram.terminal_symbols); gram.terminal_symbols = NULL;

    for (l = gram.rules; l != NULL; l = l->next) {
	Rule *rule = l->data;
	delete_rule(rule);
    }
    g_slist_free(gram.rules); gram.rules = NULL;

    clear_other_data();
} /* clear_all_data */

/*
 *  Needed for tables.c (info display)
 */

gchar *get_fin_aut_description(gboolean printable)
{
    GSList *l;
    GString *s;
    gchar *ret;

    s = g_string_sized_new(20);
    if (printable)
	s = g_string_append(s, "(T, Q, $\\delta$, \\{ ");
    else
	s = g_string_append(s, "(T, Q, d, { ");

    for (l = aut.states; l != NULL; l = l->next) {
	State *state = l->data;
	if (state->atribute != NULL && strstr(state->atribute, "s")) {
	    s = g_string_append(s, state->name);
	    break;
	}
    }
    if (printable)
	s = g_string_append(s, " \\}, F)");
    else
	s = g_string_append(s, " }, F)");

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;
} /* get_fin_aut_description */

gchar *get_stack_aut_description(gboolean printable)
{
    GSList *l;
    GString *s;
    gchar *ret;

    s = g_string_sized_new(30);
    if (printable)
	s = g_string_append(s, "(T, Q, Z, $\\delta$, \\{ ");
    else
	s = g_string_append(s, "(T, Q, Z, d, { ");

    for (l = aut.states; l != NULL; l = l->next) {
	State *state = l->data;
	if (state->atribute != NULL && strstr(state->atribute, "s")) {
	    s = g_string_append(s, state->name);
	    break;
	}
    }
    if (printable)
	s = g_string_append(s, " \\}, \\{ ");
    else
	s = g_string_append(s, " }, { ");

    for (l = aut.stack_symbols; l != NULL; l = l->next) {
	Symbol *symbol = l->data;
	if (symbol->atribute != NULL && strstr(symbol->atribute, "s")) {
	    s = g_string_append(s, symbol->name);
	    break;
	}
    }
    if (printable)
	s = g_string_append(s, " \\}, F)");
    else
	s = g_string_append(s, " }, F)");

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;
} /* get_stack_aut_description */

gchar *get_string_of_states(void)
{
    GSList *l;
    GString *s;
    gchar *ret;

    s = g_string_sized_new(10);
    s = g_string_append(s, "{ ");

    for (l = aut.states; l != NULL; l = l->next) {
	State *state = l->data;
	s = g_string_append(s, state->name);
	if (l->next != NULL)
	    s = g_string_append(s, ", ");
    }
    s = g_string_append(s, " }");

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;
} /* get_string_of_states */

gchar *get_string_of_ending_states(void)
{
    GSList *l;
    GString *s;
    gchar *ret;

    s = g_string_sized_new(10);
    s = g_string_append(s, "{ ");

    for (l = aut.states; l != NULL; l = l->next) {
	State *state = l->data;
	if (state->atribute != NULL && strstr(state->atribute, "e")) {
	    if (s->len > 2)
		s = g_string_append(s, ", ");
	    s = g_string_append(s, state->name);
	}
    }
    s = g_string_append(s, " }");

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;
} /* get_string_of_ending_states */

gchar *get_gram_description(gboolean printable)
{
    GSList *l;
    GString *s;
    gchar *ret;

    s = g_string_sized_new(15);
    if (printable)
	s = g_string_append(s, "(N, T, P, \\{ ");
    else
	s = g_string_append(s, "(N, T, P, { ");

    for (l = gram.nonterminal_symbols; l != NULL; l = l->next) {
	Symbol *symbol = l->data;
	if (symbol->atribute != NULL && strstr(symbol->atribute, "s")) {
	    s = g_string_append(s, symbol->name);
	    break;
	}
    }
    if (printable)
	s = g_string_append(s, " \\})");
    else
	s = g_string_append(s, " })");

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;

} /* get_gram_description */

gchar *get_string_of_symbols(GSList *list, gboolean set)
{
    GSList *l;
    GString *s;
    gchar *ret;

    s = g_string_sized_new(10);
    if (set)
	s = g_string_append(s, "{ ");

    for (l = list; l != NULL; l = l->next) {
	Symbol *symbol = l->data;
	s = g_string_append(s, symbol->name);
	if (l->next != NULL) {
	    if (set)
		s = g_string_append(s, ", ");
	    else
		s = g_string_append_c(s, ' ');
	}
    }
    if (set)
	s = g_string_append(s, " }");

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;
} /* get_string_of_symbols */

/*
 * Needed for graphics.c (finite automata only)
 */

gchar *get_string_of_entry_symbols(Transition *trans)
{
    gchar *ret;
    GSList *l;
    GString *symbols;

    symbols = g_string_sized_new(10);

    for (l = trans->entry_symbols; l != NULL; l = l->next) {
	Symbol *symbol = l->data;
	symbols = g_string_append(symbols, symbol->name);
	if (l->next != NULL)
	    symbols = g_string_append(symbols, " | ");
    }

    ret = g_strdup(symbols->str);
    g_string_free(symbols, TRUE);
    return ret;
} /* get_string_of_symbols */

/*
 * Needed for gui.c (input from entry boxes)
 */

gchar *get_id_string_of_symbols(gchar *sentence, GSList *list1, GSList *list2)
{
    GSList *l;
    GString *s;
    gchar *ret;
    gchar **words;
    int i = 0;

    words = g_strsplit(sentence, " ", 0);

    s = g_string_sized_new(10);

    while ((ret = words[i++]) != NULL) {
	if (strcmp(ret, "") == 0)
	    continue;
	for (l = list1; l != NULL; l = l->next) {
	    Symbol *symbol = l->data;
	    if (strcmp(ret, symbol->name) == 0) {
		g_string_sprintfa(s, "%i ", symbol->id);
		break;
	    }
	}
	if (l == NULL && list2 != NULL)
	    for (l = list2; l != NULL; l = l->next) {
		Symbol *symbol = l->data;
		if (strcmp(ret, symbol->name) == 0) {
		    g_string_sprintfa(s, "%i ", symbol->id);
		    break;
		}
	    }
	if (l == NULL) {
	    g_string_free(s, TRUE);
	    g_strfreev(words);
	    return NULL;
	}
    }
    ret = g_strdup(s->str);
    ret = g_strchomp(ret);
    g_string_free(s, TRUE);
    g_strfreev(words);
    return ret;
} /* get_id_string_of_symbols */

gchar *deparen(gchar *text)
{
    GString *s = g_string_sized_new(10);
    gchar *ret;
    int i;

    for (i = 0; i < strlen(text); i++) {
	if (text[i] == '(' || text[i] == ')')
	    s = g_string_append_c(s, '\007');
	s = g_string_append_c(s, text[i]);
    }

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;
} /* deparen */

gchar *debrack(gchar *text)
{
    GString *s = g_string_sized_new(10);
    gchar *ret;
    int i;

    for (i = 0; i < strlen(text); i++) {
	if (text[i] == '{' || text[i] == '}')
	    s = g_string_append_c(s, '\\');
	s = g_string_append_c(s, text[i]);
    }

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;
} /* debrack */

gchar *delatex(gchar *text)
{
    GString *s = g_string_sized_new(10);
    gchar *ret;
    int i;

    if (strcmp(text, "") == 0) {
	s = g_string_append(s, "$\\epsilon$");
	ret = g_strdup(s->str);
	g_string_free(s, TRUE);
	return ret;
    }

    for (i = 0; i < strlen(text); i++) {
	if (text[i] == '{' || text[i] == '}')
	    s = g_string_append_c(s, '\\');
	s = g_string_append_c(s, text[i]);
    }

    ret = g_strdup(s->str);
    g_string_free(s, TRUE);
    return ret;
} /* delatex */

gboolean bidirectional(Transition *trans)
{
    GSList *l;
    for (l = aut.transitions; l != NULL; l = l->next) {
	Transition *t = l->data;
	if (trans->from == t->to && trans->to == t->from)
	    return TRUE;
    }
    return FALSE;
} /* bidirect */

void draw_all_states()
{
    GSList *l;

    for (l = aut.states; l != NULL; l = l->next) {
	State *state = l->data;
	if (state->group == NULL) {
	    double x = 10.0;
	    double y = 10.0;

	    Coordinate *coordinate = find_coordinate_by_id(state->id);
	    if (coordinate != NULL) {
		x = coordinate->x;
		y = coordinate->y;
	    } else
		add_coordinate(state->id, x, y); /*TIP: not needed*/
	    state->group = create_circle_with_text(state->name, x, y,
						   state->atribute);
	}
    }
} /* draw_all_states */

void draw_all_transitions()
{
    GSList *l;
    gchar *symbols;

    for (l = aut.transitions; l != NULL; l = l->next) {
	Transition *trans = l->data;
	if (trans->group != NULL)
	    gtk_object_destroy(GTK_OBJECT(trans->group));

	symbols = get_string_of_entry_symbols(trans);
	if (bidirectional(trans))
	    trans->group = create_arcarrow_with_text(trans->from, trans->to,
						     symbols);
	else
	    trans->group = create_arrow_with_text(trans->from, trans->to,
						  symbols);
	g_free(symbols);
    }
} /* draw_all_transitions */
