/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Miroslav Kure <kurem@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gnome.h>
#include <libgnomeui/gnome-canvas.h>
#include <stdlib.h>
#include <math.h>
#include "data.h"
#include "functions.h"
#include "gui.h"
#include "table.h"
#include "graphics.h"
#include "popup.h"

static GnomeCanvas *canvas;
static GnomeCanvasGroup *drawing_area;
static gint transitioning;
char *font;

void prepare_transition()
{
    clear_other_data();
    transitioning = 2;
} /* prepare_transition */

static void redraw_moving(GnomeCanvasItem *item)
{
    GSList *l;
    State *state;
    gchar *symbols;

    state = find_state_by_group(GNOME_CANVAS_GROUP(item));
    g_return_if_fail(state != NULL);

    l = aut.transitions;
    while (l != NULL) {
	Transition *trans = l->data;
	if (trans->from == state || trans->to == state) {
	    gtk_object_destroy(GTK_OBJECT(trans->group));
	    trans->group = NULL;
	    symbols = get_string_of_entry_symbols(trans);
	    if (bidirectional(trans))
		trans->group = create_arcarrow_with_text(trans->from,
							 trans->to,
							 symbols);
	    else
		trans->group = create_arrow_with_text(trans->from,
						      trans->to,
						      symbols);
	    g_free(symbols);
	}
	l = l->next;
    }
} /* redraw_moving */

static gint
background_event(GnomeCanvasItem *item, GdkEventButton *event, gpointer data)
{
    if (event->type == GDK_BUTTON_PRESS)
	transitioning = 0;
    if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
	Coordinate *coord = g_new(Coordinate, 1);
	coord->id = 0;
	coord->x = event->x;
	coord->y = event->y;

	gnome_popup_menu_do_popup(gnome_popup_menu_new(popup),
				  NULL, NULL, NULL, coord);
	return TRUE;
    }
    return FALSE;
} /* background_event */

static gint
state_event(GnomeCanvasItem *item, GdkEventButton *event, gpointer data)
{
    static double x, y;
    double new_x, new_y;
    GdkCursor *fleur;
    static int dragging;
    double item_x, item_y;

    item_x = event->x;
    item_y = event->y;

    gnome_canvas_item_w2i(item->parent, &item_x, &item_y);

    switch (event->type) {

    case GDK_BUTTON_PRESS:
	if (event->button == 1) {
	    if (transitioning) {
		State *state = find_state_by_group(GNOME_CANVAS_GROUP(item));
		g_return_val_if_fail(state != NULL, FALSE);
		add_state_to_list(state->id);

		if (--transitioning == 0)
		    direct_add_transition_cb(GTK_WIDGET(canvas), NULL);
		return TRUE;
	    }

	    x = item_x;
	    y = item_y;

	    fleur = gdk_cursor_new(GDK_FLEUR);
	    gnome_canvas_item_grab(item,
				   GDK_POINTER_MOTION_MASK | 
				   GDK_BUTTON_RELEASE_MASK,
				   fleur,
				   event->time);
	    gdk_cursor_destroy(fleur);
	    dragging = TRUE;
	    gnome_canvas_item_raise_to_top(item);
	} /* left 'moving' button */

	else if (event->button == 3) {
	    gnome_popup_menu_do_popup(gnome_popup_menu_new(state_popup),
				      NULL, NULL, NULL, item);
	    return TRUE;
	} /* right 'option'  button */
	break;
	
    case GDK_MOTION_NOTIFY:
	if (dragging && (event->state & GDK_BUTTON1_MASK)) {
	    new_x = (item_x < 0)? 0 : item_x;
	    new_y = (item_y < 0)? 0 : item_y;
            
	    gnome_canvas_item_move(item, new_x - x, new_y - y);
	    x = new_x;
	    y = new_y;
	    redraw_moving(item);
	}
	break;
	
    case GDK_BUTTON_RELEASE:
    {
	if (transitioning)
	    return TRUE;
	else {
	    Coordinate *coordinate;
	    State *state = find_state_by_group(GNOME_CANVAS_GROUP(item));
	    g_return_val_if_fail(state != NULL, FALSE);
	    coordinate = find_coordinate_by_id(state->id);
	    g_return_val_if_fail(coordinate != NULL, FALSE);
	    coordinate->x = x;
	    coordinate->y = y;
	}
	gnome_canvas_item_ungrab(item, event->time);
	dragging = FALSE;
	break;
    }
    default:
	break;
    }
    return TRUE;
} /* state_event */

static gint
transition_event(GnomeCanvasItem *item, GdkEventButton *event, gpointer data)
{
    if (event->type == GDK_BUTTON_PRESS && event->button == 3 ) {
	gnome_popup_menu_do_popup(gnome_popup_menu_new(transition_popup),
				  NULL, NULL, NULL, item);
	return TRUE;
    }
    return FALSE;
} /* transition_event */


GnomeCanvasGroup *
create_circle_with_text(gchar *name, double x, double y, gchar *atribute)
{
    GnomeCanvasGroup *group;
    GnomeCanvasItem *text, *circle;
    double width, height;

    group = GNOME_CANVAS_GROUP(
	gnome_canvas_item_new(drawing_area,
			      gnome_canvas_group_get_type(),
			      "x", x,
			      "y", y,
			      NULL));

    text = gnome_canvas_item_new(group,
			  gnome_canvas_text_get_type(),
			  "text", name,
			  "x", TEXT_OFFSET +BORDER_OFFSET,
			  "y", TEXT_OFFSET/2 +BORDER_OFFSET,
			  "font", font,
			  "anchor", GTK_ANCHOR_NORTH_WEST,
			  "fill_color_rgba", 0x33111111,
			  NULL);
    gtk_object_get(GTK_OBJECT(text), "text_width", &width, NULL);
    gtk_object_get(GTK_OBJECT(text), "text_height", &height, NULL);

    if (atribute != NULL && strstr(atribute, "e"))
	gnome_canvas_item_new(group,
			      gnome_canvas_ellipse_get_type(),
			      "x1", 0.0,
			      "y1", 0.0,
			      "x2", width + 2*TEXT_OFFSET +2*BORDER_OFFSET,
			      "y2", height +TEXT_OFFSET +2*BORDER_OFFSET,
			      "fill_color_rgba", 0xeeeec700,
			      "outline_color", "black",
			      "width_pixels", 1,
			      NULL);

    if (atribute != NULL && strstr(atribute, "s"))
	gnome_canvas_item_new(group,
			      gnome_canvas_ellipse_get_type(),
			      "x1", -2*BORDER_OFFSET,
			      "y1", TEXT_OFFSET +2.0,
			      "x2", -BORDER_OFFSET,
			      "y2", TEXT_OFFSET +BORDER_OFFSET +2.0,
			      "fill_color_rgba", 0x11119900,
			      "outline_color", "black",
			      "width_pixels", 1,
			      NULL);

    circle = gnome_canvas_item_new(group,
				   gnome_canvas_ellipse_get_type(),
				   "x1", BORDER_OFFSET,
				   "y1", BORDER_OFFSET,
				   "x2", width + 2*TEXT_OFFSET +BORDER_OFFSET,
				   "y2", height +TEXT_OFFSET +BORDER_OFFSET,
				   "fill_color_rgba", 0xeeeec700,
				   "outline_color", "black",
				   "width_pixels", 1,
				   NULL);
    gtk_object_set_data(GTK_OBJECT(group), "circle", circle);
    gnome_canvas_item_raise_to_top(text);

    gtk_signal_connect(GTK_OBJECT(group),
		       "event",
		       (GtkSignalFunc) state_event,
		       NULL);
    return group;
} /*create_circle_with_text */

GnomeCanvasGroup *create_arrow_with_text(State *from, State *to, char *text)
{
    GnomeCanvasGroup *group;
    GnomeCanvasPoints *points;
    double x, y, ux, uy;
    double angle, s_angle, pi_6, step_a;
    double fx, fy, tx, ty, d, ax, ay;
    double fx1, fy1, fx2, fy2;
    double tx1, ty1, tx2, ty2;
    int i, n;
    GtkAnchorType anchor_type;

    gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(from->group),
				 &fx1, &fy1, &fx2, &fy2);
    gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(to->group),
				 &tx1, &ty1, &tx2, &ty2);

    fx = (fx1+fx2)/2;
    fy = (fy1+fy2)/2;
    tx = (tx1+tx2)/2;
    ty = (ty1+ty2)/2;

    if (fx1 < tx1)
	x = fx1;
    else
	x = tx1;

    if (fy1 < ty1)
	y = fy1;
    else
	y = ty1;
    
    ux = tx - fx;
    uy = ty - fy;

    angle = acos(ux / sqrt(ux*ux + uy*uy));

    if (from == to) {
	n = 20;
	points = gnome_canvas_points_new(n);
	d = 20.0;
	pi_6 = (2.0/3.0)*atan(1); step_a = 12*pi_6/(n);
	s_angle = -0.3;
	for (i = 0; i < n; ++i, s_angle += step_a) {
	    ax = d*cos(s_angle); ay = -d*sin(s_angle);
	    points->coords[2*i] = ax + fx-x-(fx-fx1);
	    points->coords[2*i+1] = ay + fy-y;
	}
	ax = -(fx-fx1)-20; ay = -20;
    } else {
	points = gnome_canvas_points_new(2);
	points->coords[0] = fx-x;
	points->coords[1] = fy-y;
	points->coords[2] = tx-x - (tx2-tx1)*cos(angle)/2;
	if (fy1 < ty1)
	  points->coords[3] = ty-y - (ty2-ty1)*sin(angle)/2;
	else
	  points->coords[3] = ty-y + (ty2-ty1)*sin(angle)/2;
	ax = ay = 0;
    }

    group = GNOME_CANVAS_GROUP(
	gnome_canvas_item_new(drawing_area,
			      gnome_canvas_group_get_type(),
			      "x", x,
			      "y", y,
			      NULL));

    gnome_canvas_item_new(group,
			  gnome_canvas_line_get_type (),
			  "points", points,
			  "fill_color", "brown",
			  "width_pixels", 2,
			  "last_arrowhead", TRUE,
			  "arrow_shape_a", 5.0,
			  "arrow_shape_b", 10.0,
			  "arrow_shape_c", 5.0,
			  NULL);

    if (fx1 < tx1) { 
	if (fy1 < ty1) anchor_type = GTK_ANCHOR_SW; 
	else anchor_type = GTK_ANCHOR_NW;
    }
    else { 
	if (fy1 < ty1) anchor_type = GTK_ANCHOR_SE; 
	else anchor_type = GTK_ANCHOR_NE;
    }

    gnome_canvas_item_new(group,
			  gnome_canvas_text_get_type(),
			  "text", text,
			  "x", ax +(tx+fx)/2 -x,
			  "y", ay +(ty+fy)/2 -y,
			  "font", font,
			  "anchor", anchor_type,
			  "fill_color_rgba", 0x66644488,
			  NULL);

    gnome_canvas_item_lower_to_bottom(GNOME_CANVAS_ITEM(group));

    gtk_signal_connect(GTK_OBJECT(group),
		       "event",
		       (GtkSignalFunc) transition_event,
		       NULL);
    gnome_canvas_points_free(points);

    return group;
} /* create_arrow_with_text */

GnomeCanvasGroup *create_arcarrow_with_text(State *from, State *to, char *text)
{
    GnomeCanvasGroup *group;
    GnomeCanvasPoints *points;
    double x, y, ux, uy;
    double s_angle, step_a, pi_6;
    double fx, fy, tx, ty, d, dx, dy, s_, c_, sq_, step_x, ax, ay, d_;
    double fx1, fy1, fx2, fy2;
    double tx1, ty1, tx2, ty2;
    int i, n;
    GtkAnchorType anchor_type;
     
    gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(from->group),
				 &fx1, &fy1, &fx2, &fy2);
    gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(to->group),
				 &tx1, &ty1, &tx2, &ty2);
 
    fx = (fx1+fx2)/2;
    fy = (fy1+fy2)/2;
    tx = (tx1+tx2)/2;
    ty = (ty1+ty2)/2;
    
    if (fx1 < tx1)
	x = fx1;
    else
	x = tx1;
 
    if (fy1 < ty1)
	y = fy1;
    else
	y = ty1;
     
    ux = tx - fx;
    uy = ty - fy;
 
    c_ = ux / sqrt(ux*ux + uy*uy); s_ = uy / sqrt(ux*ux + uy*uy);
 
    if (from == to) {
	n = 20;
	points = gnome_canvas_points_new(n);
	d = 20.0;
	pi_6 = (2.0/3.0)*atan(1); step_a = 12*pi_6/(n);
	s_angle = -0.25;
	for (i = 0; i < n; ++i, s_angle += step_a) {
	    ax = d*cos(s_angle); ay = -d*sin(s_angle);
	    points->coords[2*i] = ax + fx-x-(fx-fx1);
	    points->coords[2*i+1] = ay + fy-y;
	}

	ax = -(fx-fx1)-20; ay = -10;
    } else {
	n = 9;
	points = gnome_canvas_points_new(n+2);
	sq_ = sqrt(3)/2;
	d = sqrt(ux*ux + uy*uy);
	d_ = sqrt(((tx2-tx1)*c_/2)*((tx2-tx1)*c_/2) + 
		  ((ty2-ty1)*s_/2)*((ty2-ty1)*s_/2));
	pi_6 = (2.0/3.0)*atan(1); step_a = 2*pi_6/(n+1);
	s_angle = -pi_6 + step_a;
	step_x = d / (n+1); dx = step_x;
	points->coords[0] = fx-x;
	points->coords[1] = fy-y;
	for (i = 1; i < n+2; ++i, s_angle += step_a, dx += step_x) {
	    if (i == n+1) { dx = d-d_+2.0; s_angle = -pi_6 + 2*pi_6*dx/d; }
	    dy = -0.5*d*(cos(s_angle)-sq_);
	    ax = dx*c_ - dy*s_; ay = dx*s_ + dy*c_;
	    points->coords[2*i] = ax + fx-x;
	    points->coords[2*i+1] = ay + fy-y;
	    if (dx > d-d_-step_x) { s_angle -= step_a; dx -= step_x; }
	}
	dx = d/2; dy = -0.5*d*(1-sq_);
	ax = dx*c_ - dy*s_; ay = dx*s_ + dy*c_;
    }

    group = GNOME_CANVAS_GROUP(
	gnome_canvas_item_new(drawing_area,
			      gnome_canvas_group_get_type(),
			      "x", x,
			      "y", y,
			      NULL));
 
    gnome_canvas_item_new(group,
			  gnome_canvas_line_get_type (),
			  "points", points,
			  "fill_color", "brown",
			  "width_pixels", 2,
			  "last_arrowhead", TRUE,
			  "arrow_shape_a", 5.0,
			  "arrow_shape_b", 10.0,
			  "arrow_shape_c", 5.0,
			  NULL);
 
    if (fx1 < tx1) { 
	if (fy1 < ty1) anchor_type = GTK_ANCHOR_NW; 
	else anchor_type = GTK_ANCHOR_SW;
    }
    else { 
	if (fy1 < ty1) anchor_type = GTK_ANCHOR_NE; 
	else anchor_type = GTK_ANCHOR_SE;
    }
 
    gnome_canvas_item_new(group,
			  gnome_canvas_text_get_type(),
			  "text", text,
			  "x", ax + fx-x,
			  "y", ay + fy-y,
			  "font", font,
			  "anchor", anchor_type,
			  "fill_color_rgba", 0x66644488,
			  NULL);
 
    gnome_canvas_item_lower_to_bottom(GNOME_CANVAS_ITEM(group));
 
    gtk_signal_connect(GTK_OBJECT(group),
		       "event",
		       (GtkSignalFunc) transition_event,
		       NULL);
    gnome_canvas_points_free(points);
 
    return group;
} /* create_arcarrow_with_text */

void show_finite_automata_frame(void)
{
    GtkWidget *scroll;
    GtkWidget *c;
    GnomeCanvasItem *back;

    gtk_widget_destroy(main_vbox);

    transitioning = 0;
    if (font == NULL)
	font = g_strdup("fixed");

    gtk_widget_push_visual(gdk_imlib_get_visual());
    gtk_widget_push_colormap(gdk_imlib_get_colormap());

    c = gnome_canvas_new();
    canvas = GNOME_CANVAS(c);

    gtk_widget_pop_visual();
    gtk_widget_pop_colormap();

    back = gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(canvas)),
				 gnome_canvas_rect_get_type(),
				 "outline_color", "black",
				 "x1", 0.0,
				 "y1", 0.0,
				 "x2", 1600.0,
				 "y2", 1400.0,
				 "fill_color", "white",
				 "width_units", 0.0,
				 NULL);

    gtk_signal_connect(GTK_OBJECT(back), "event",
		       (GtkSignalFunc) background_event, NULL);
    
    gnome_canvas_item_lower_to_bottom(back);

    drawing_area = GNOME_CANVAS_GROUP(
	gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(canvas)),
			      gnome_canvas_group_get_type(),
			      "x", 0.0,
			      "y", 0.0,
			      NULL));
    
    scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
				    GTK_POLICY_AUTOMATIC,
				    GTK_POLICY_AUTOMATIC);
    gnome_canvas_set_scroll_region(canvas, 0.0, 0.0, 1600.0, 1400.0);
    gtk_container_add(GTK_CONTAINER(scroll), c);

    draw_all_states();
    draw_all_transitions();
    if (states != NULL) {
	State *state = states->data;
	if (state != NULL) {
	    GnomeCanvasItem *circle = gtk_object_get_data(
		GTK_OBJECT(state->group), "circle");
	    gtk_object_set(GTK_OBJECT(circle),
			   "fill_color_rgba", 0xeeaaaa00, NULL);
	}
    }

    main_vbox = gtk_vbox_new (FALSE, 0);
    make_finite_automata_table(main_vbox);
    gtk_box_pack_start(GTK_BOX(main_vbox), scroll, TRUE, TRUE, 0);

    gnome_app_set_contents(GNOME_APP(app), main_vbox);
    gtk_widget_show_all (app);
}  /* show_finite_automata_frame */


