/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "error.h"
#include "automata.h"

Automata::Automata()
{
  start_state = State_NULL;
  c_state = State_NULL;
};

Automata::~Automata()
{

};

State
Automata::addState()
{
  State new_state = State_MIN;
  for (set<State>::const_iterator iter_st = states.begin();
       iter_st != states.end(); ++iter_st, ++new_state)
    if (*iter_st != new_state) break;
  if (new_state > State_MAX) { setError(E_A_STATE_MAX); return State_NULL; }
  states.insert(new_state);
  // if there's only one state it's start and configuration
  if (states.size() == 1) { start_state = new_state; c_state = start_state; }
  return new_state;
}

bool
Automata::removeState(State state)
{
  set<State>::iterator iter_st_f = states.find(state);
  if (iter_st_f == states.end()) { setError(E_A_STATE_BAD); return false; }
  // erase also transitions which contain the state
  set<Transition> trans = find_transitions_with_state(state);
  for (set<Transition>::const_iterator iter_t = trans.begin();
       iter_t != trans.end(); ++iter_t) removeTransition(*iter_t);
  // if we're erasing start state we must set new one
  if (state == start_state) {
    if (states.size() != 1) {
      if (state == *states.begin()) start_state = *++states.begin();
      else start_state = *states.begin();
    }
    else start_state = State_NULL;
  }
  // if we're erasing configuration state we must set new one (start)
  if (state == c_state) c_state = start_state;
  states.erase(state);
  return true;
}

set<Transition>
Automata::find_transitions_with_state(State state)
{
  set<Transition> result;
  for (set<Transition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t) {
    if (iter_t->from_state == state) { result.insert(*iter_t); continue; }
    if (iter_t->to_state == state) result.insert(*iter_t);
  }
  return result;
}

inline set<State>
Automata::getStates()
{
  return states;
}

EntrySymbol
Automata::addEntrySymbol()
{
  EntrySymbol new_symbol = EntrySymbol_MIN;
  for (set<EntrySymbol>::const_iterator iter_sy = entry_symbols.begin();
       iter_sy != entry_symbols.end(); ++iter_sy, ++new_symbol)
    if (*iter_sy != new_symbol) break;
  if (new_symbol > EntrySymbol_MAX) { 
    setError(E_A_ENTRY_SYMBOL_MAX); return EntrySymbol_NULL;
  }
  entry_symbols.insert(new_symbol);
  return new_symbol;
}

bool
Automata::removeEntrySymbol(EntrySymbol symbol)
{
  set<EntrySymbol>::iterator iter_sy_f = entry_symbols.find(symbol);
  if (iter_sy_f == entry_symbols.end()) { 
    setError(E_A_ENTRY_SYMBOL_BAD); return false;
  }
  // erase also transitions which contain the symbol
  set<Transition> trans = find_transitions_with_entry_symbol(symbol);
  for (set<Transition>::const_iterator iter_t = trans.begin();
       iter_t != trans.end(); ++iter_t) removeTransition(*iter_t);
  entry_symbols.erase(symbol);
  return true;
}

set<Transition>
Automata::find_transitions_with_entry_symbol(EntrySymbol symbol)
{
  set<Transition> result;
  for (set<Transition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t)
    if (iter_t->entry_symbol == symbol) result.insert(*iter_t);
  return result;
}

inline set<EntrySymbol>
Automata::getEntrySymbols()
{
  return entry_symbols;
}

bool
Automata::removeTransition(Transition transition)
{
  set<Transition>::iterator iter_t_f = transitions.find(transition);
  if (iter_t_f == transitions.end()) { 
    setError(E_A_TRANSITION_BAD); return false;
  }
  transitions.erase(transition);
  return true;
}

inline set<Transition>
Automata::getTransitions()
{
  return transitions;
}

bool
Automata::setStartState(State state)
{
  // check if the state is in the states set
  set<State>::iterator iter_st_f = states.find(state);
  if (iter_st_f == states.end()) { setError(E_A_STATE_BAD); return false; }
  start_state = state;
  return true;
}

inline State
Automata::getStartState()
{
  return start_state;
}

bool
Automata::setEndState(State state)
{
  // check if the state is in the states set
  set<State>::iterator iter_st_f = states.find(state);
  if (iter_st_f == states.end()) { setError(E_A_STATE_BAD); return false; }
  end_states.insert(state);
  return true;
}

bool
Automata::resetEndState(State state)
{
  // check if the state is in the end_states set
  set<State>::iterator iter_st_f = end_states.find(state);
  if (iter_st_f == end_states.end()) { setError(E_A_STATE_BAD); return false; }
  end_states.erase(state);
  return true;
}

inline set<State>
Automata::getEndStates()
{
  return end_states;
}
