/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "error.h"
#include "contextfree_grammar.h"

ContextfreeGrammar::ContextfreeGrammar()
{

}

ContextfreeGrammar::~ContextfreeGrammar()
{

}

bool
ContextfreeGrammar::addRule(Rule &rule, vector<Symbol> left,
			    vector<Symbol> right)
{
  // check if all symbols in both parts are in the [non]terminal_symbols set
  // CHOICE: if not, add it - now NO
  set<Symbol>::iterator iter_sy_f;
  for (vector<Symbol>::const_iterator iter_sy = left.begin();
       iter_sy != left.end(); ++iter_sy)
    if (isSymbolNonterminal(*iter_sy)) {
      iter_sy_f = nonterminal_symbols.find(*iter_sy);
      if (iter_sy_f == nonterminal_symbols.end()) {
        setError(E_G_NONTERMINAL_SYMBOL_BAD); return false;
      }
    } else {
      iter_sy_f = terminal_symbols.find(*iter_sy);
      if (iter_sy_f == terminal_symbols.end()) {
        setError(E_G_TERMINAL_SYMBOL_BAD); return false;
      }
    }
  for (vector<Symbol>::const_iterator iter_sy = right.begin();
       iter_sy != right.end(); ++iter_sy)
    if (isSymbolNonterminal(*iter_sy)) {
      iter_sy_f = nonterminal_symbols.find(*iter_sy);
      if (iter_sy_f == nonterminal_symbols.end()) {
        setError(E_G_NONTERMINAL_SYMBOL_BAD); return false;
      }
    } else {
      iter_sy_f = terminal_symbols.find(*iter_sy);
      if (iter_sy_f == terminal_symbols.end()) {
        setError(E_G_TERMINAL_SYMBOL_BAD); return false;
      }
    }
  // check if the rule suits criterions for the contextfree grammar rule
  if (left.size() != 1) { setError(E_GC_RULE_BAD);return false; }
  if (!isSymbolNonterminal(left.front())) { 
    setError(E_GC_RULE_BAD); return false;
  }
  rule.left = left;
  rule.right = right;
  rules.insert(rule);
  return true;
}

bool
ContextfreeGrammar::addRule(vector<Symbol> left, vector<Symbol> right)
{
  Rule rule;
  return addRule(rule, left, right);
}

set<NonterminalSymbol>
ContextfreeGrammar::getRedundantSymbols()
{
  set<NonterminalSymbol> nrs;
  set<Symbol> test_set = terminal_symbols;
  vector<Symbol>::const_iterator iter_sy;
  set<Symbol>::iterator iter_sy_f;
  set<Rule> test_rules, rules_;
  bool inserted;
  test_rules = rules;
  do {
    inserted = false;
    rules_ = test_rules;
    for (set<Rule>::const_iterator iter_r = rules_.begin();
	 iter_r != rules_.end(); ++iter_r) {
      for (iter_sy = iter_r->right.begin();
	   iter_sy != iter_r->right.end(); ++iter_sy) {
	iter_sy_f = test_set.find(*iter_sy);
	if (iter_sy_f == test_set.end()) break;
      }
      if (iter_sy == iter_r->right.end()) { 
	pair<set<Symbol>::iterator, bool>
	  p = nrs.insert(iter_r->left.front());
	if (p.second) inserted = true;
	test_rules.erase(*iter_r);
      }
    }
    test_set.insert(nrs.begin(), nrs.end());
  } while (inserted);
  set<NonterminalSymbol> result;
  set_difference(nonterminal_symbols.begin(), nonterminal_symbols.end(), 
		 nrs.begin(), nrs.end(), 
		 insert_iterator<set<NonterminalSymbol> >
		 (result, result.begin()));
  return result;
}

set<Symbol>
ContextfreeGrammar::getInaccessibleSymbols()
{
  set<Symbol> nis;
  nis.insert(start_symbol);
  set<Symbol> test_set;  
  test_set.insert(start_symbol);
  set<Symbol>::iterator iter_sy_f;
  set<Rule> test_rules, rules_;
  bool inserted;
  test_rules = rules;
  do {
    inserted = false; 
    rules_ = test_rules;
    for (set<Rule>::const_iterator iter_r = rules_.begin();
	 iter_r != rules_.end(); ++iter_r) {
      iter_sy_f = test_set.find(iter_r->left.front());
      if (iter_sy_f != test_set.end()) {
	for (vector<Symbol>::const_iterator iter_sy = iter_r->right.begin();
	     iter_sy != iter_r->right.end(); ++iter_sy) { 
	  pair<set<Symbol>::iterator, bool> p = nis.insert(*iter_sy);
	  if (p.second) inserted = true;
	}
      	test_rules.erase(*iter_r);
      }
    }
    test_set.insert(nis.begin(), nis.end());
  } while (inserted);
  set<Symbol> result;
  set<Symbol> as;
  set_union(nonterminal_symbols.begin(), nonterminal_symbols.end(),
	    terminal_symbols.begin(), terminal_symbols.end(),
	    insert_iterator<set<Symbol> >(as, as.begin()));
  set_difference(as.begin(), as.end(), nis.begin(), nis.end(), 
		 insert_iterator<set<Symbol> >(result, result.begin()));
  return result;
}

void
ContextfreeGrammar::removeRedundantSymbols()
{
  set<NonterminalSymbol> rs = getRedundantSymbols();
  set<Symbol> is;
  // if start symbol is redundant, all symbols are redundant (even terminal)
  set<NonterminalSymbol>::iterator iter_sy_f = rs.find(start_symbol);
  if (iter_sy_f != rs.end()) {
    rs = nonterminal_symbols;
    is = terminal_symbols;
  }
  // remove(Non)terminalSymbol method removes also rules which contain that
  // symbol
  for (set<NonterminalSymbol>::const_iterator iter_sy = rs.begin();
       iter_sy != rs.end(); ++iter_sy)
    removeNonterminalSymbol(*iter_sy);
  if (is.empty()) is = getInaccessibleSymbols();
  for (set<Symbol>::const_iterator iter_sy = is.begin(); iter_sy != is.end();
       ++iter_sy)
    if (isSymbolNonterminal(*iter_sy)) removeNonterminalSymbol(*iter_sy);
    else removeTerminalSymbol(*iter_sy);
}

set<Rule>
ContextfreeGrammar::getERules()
{
  set<Rule> result;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r)
    if (iter_r->right.empty()) result.insert(*iter_r);
  return result;
}

bool
ContextfreeGrammar::isERulesfree()
{
  set<Rule> erules = getERules();
  if (erules.empty()) return true;
  if (erules.size() > 1) return false;
  // there can be the rule S->e as long as the start symbol S is not in the
  // right part of any rule and this rule is the only one e-rule
  vector<Symbol> str = erules.begin()->left;
  if (str.front() == start_symbol) {
    for (set<Rule>::const_iterator iter_r = rules.begin();
	 iter_r != rules.end(); ++iter_r)
      for (vector<Symbol>::const_iterator iter_sy = iter_r->right.begin();
         iter_sy != iter_r->right.end(); ++iter_sy) 
	if (*iter_sy == start_symbol) return false;
    return true;
  }
  else return false;
}

NonterminalSymbol
ContextfreeGrammar::removeERules()
{
  NonterminalSymbol result = start_symbol;
  if (isERulesfree()) return result;
  set<NonterminalSymbol> es;
  set<NonterminalSymbol> test_set;
  vector<Symbol>::const_iterator iter_sy;
  set<NonterminalSymbol>::iterator iter_sy_f;
  set<Rule> test_rules, rules_;
  bool inserted;
  test_rules = rules;
  do {
    inserted = false;
    rules_ = test_rules;
    for (set<Rule>::const_iterator iter_r = rules_.begin();
	 iter_r != rules_.end(); ++iter_r) {
      for (iter_sy = iter_r->right.begin();
	   iter_sy != iter_r->right.end(); ++iter_sy) {
	iter_sy_f = test_set.find(*iter_sy);
	if (iter_sy_f == test_set.end()) break;
      }
      if (iter_sy == iter_r->right.end()) { 
	pair<set<NonterminalSymbol>::iterator, bool> p = 
	  es.insert(iter_r->left.front());
	if (p.second) inserted = true;
	test_rules.erase(*iter_r);
      }
    }
    test_set.insert(es.begin(), es.end());
  } while (inserted);
  set<Rule> new_rules, add_rules;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) {
    add_rules = modifyRule(*iter_r, es);
    // no e-rules and A->A like rules
    for (set<Rule>::const_iterator iter_r1 = add_rules.begin();
	 iter_r1 != add_rules.end(); ++iter_r1)
      if (!((iter_r1->right.empty()) || 
	    ((iter_r1->right.size() == 1) && 
	     (iter_r1->left.front() == iter_r1->right.front()))))
	new_rules.insert(*iter_r1);
  }
  rules = new_rules;
  iter_sy_f = es.find(start_symbol);
  if (iter_sy_f != es.end()) toExtended();
  return result;
}

set<Rule>
ContextfreeGrammar::modifyRule(Rule rule, set<NonterminalSymbol> ns)
{
  set<Rule> result;
  // insert original rule
  result.insert(rule);
  set<Symbol>::iterator iter_sy_f;
  set<Rule> new_rules;
  Rule new_rule = rule;
  vector<Symbol>::iterator last_iter_sy = new_rule.right.begin();
  bool erased;
  do {
    erased = false; new_rule = rule;
    for (vector<Symbol>::iterator iter_sy = new_rule.right.begin();
	 iter_sy != new_rule.right.end(); ++iter_sy)
      if (iter_sy >= last_iter_sy) {
	iter_sy_f = ns.find(*iter_sy);
	if (iter_sy_f != ns.end()) {
	  // remove nonterminal symbol if it's in the ns set
	  last_iter_sy = iter_sy; new_rule.right.erase(iter_sy); erased = true;
	  new_rules = modifyRule(new_rule, ns);
	  result.insert(new_rules.begin(), new_rules.end());
	  break;
	}
      }
    ++last_iter_sy;
  } while (erased);
  return result;
}

set<Rule>
ContextfreeGrammar::getSimpleRules()
{
  set<Rule> result;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r)
    if ((iter_r->right.size() == 1) && 
	(isSymbolNonterminal(iter_r->right.front())))
      result.insert(*iter_r);
  return result;
}

void
ContextfreeGrammar::removeSimpleRules()
{
  if (getSimpleRules().empty()) return;
  // grammar must be without e-rules
  if (!isERulesfree()) { setError(E_GC_NOT_ERULES_FREE); return; }
  map<NonterminalSymbol, set<NonterminalSymbol> > sr_map;
  set<NonterminalSymbol> test_set;
  set<NonterminalSymbol>::iterator iter_sy_f;
  set<Rule> test_rules, rules_;
  bool inserted;
  for (set<NonterminalSymbol>::const_iterator
	 iter_sy = nonterminal_symbols.begin();
       iter_sy != nonterminal_symbols.end(); ++iter_sy) {
    sr_map[*iter_sy].insert(*iter_sy);
    test_set.clear(); test_set.insert(*iter_sy);
    test_rules = rules;
    do {
      inserted = false; 
      rules_ = test_rules;
      for (set<Rule>::const_iterator iter_r = rules_.begin();
	   iter_r != rules_.end(); ++iter_r) {
	iter_sy_f = test_set.find(iter_r->left.front());
	if ((iter_sy_f != test_set.end()) &&
	    (iter_r->right.size() == 1) &&
	    (isSymbolNonterminal(iter_r->right.front()))) {
	  pair<set<NonterminalSymbol>::iterator, bool> 
	    p = sr_map[*iter_sy].insert(iter_r->right.front());
	  if (p.second) inserted = true;
	  test_rules.erase(*iter_r);
	}
      }
      test_set.insert(sr_map[*iter_sy].begin(), sr_map[*iter_sy].end());
    } while (inserted);
  }
  set<Rule> new_rules;
  Rule new_rule;
  for (set<NonterminalSymbol>::const_iterator 
	 iter_sy = nonterminal_symbols.begin();
       iter_sy != nonterminal_symbols.end(); ++iter_sy)
    for (set<NonterminalSymbol>::const_iterator 
	   iter_sy1 = sr_map[*iter_sy].begin();
	 iter_sy1 != sr_map[*iter_sy].end(); ++iter_sy1)
      for (set<Rule>::const_iterator iter_r = rules.begin();
	   iter_r != rules.end(); ++iter_r)
	if ((iter_r->left.front() == *iter_sy1) &&
	    // and not simple rules
	    (!((iter_r->right.size() == 1) &&
	       (isSymbolNonterminal(iter_r->right.front()))))) {
	  new_rule.left = vector<NonterminalSymbol>(1, *iter_sy);
	  new_rule.right = iter_r->right;
	  new_rules.insert(new_rule);
	}
  rules = new_rules;
}

bool
ContextfreeGrammar::isSelf()
{
  // grammar must be without redundant symbols (all, thus also without
  // inaccessible symbols), e-rules and cycles
  if ((!(getRedundantSymbols().empty())) || 
      (!(getInaccessibleSymbols().empty())) || (!(isERulesfree())) ||
      (hasCycle()))
    return false;
  return true;
}

bool
ContextfreeGrammar::hasCycle()
{
  map<NonterminalSymbol, set<NonterminalSymbol> > sr_map;
  set<NonterminalSymbol>::iterator iter_sy_f;
  bool inserted;
  set<NonterminalSymbol> test_set;
  set<Rule> test_rules, rules_;
  for (set<NonterminalSymbol>::const_iterator
	 iter_sy = nonterminal_symbols.begin();
       iter_sy != nonterminal_symbols.end(); ++iter_sy) {
    sr_map[*iter_sy].insert(*iter_sy);
    test_set.clear(); test_set.insert(*iter_sy);
    test_rules = rules;
    do {
      inserted = false; 
      rules_ = test_rules;
      for (set<Rule>::const_iterator iter_r = rules_.begin();
	   iter_r != rules_.end(); ++iter_r) {
	iter_sy_f = test_set.find(iter_r->left.front());
	if ((iter_sy_f != test_set.end()) && (iter_r->right.size() == 1) &&
	    (isSymbolNonterminal(iter_r->right.front()))) {
	  pair<set<NonterminalSymbol>::iterator, bool> 
	    p = sr_map[*iter_sy].insert(iter_r->right.front());
	  if (p.second) inserted = true;
	  if (iter_r->right.front() == *iter_sy) return true; //cycle
	  test_rules.erase(*iter_r);
	}
      }
      test_set.insert(sr_map[*iter_sy].begin(), sr_map[*iter_sy].end());
    } while (inserted);
  }
  return false;
}

NonterminalSymbol
ContextfreeGrammar::toSelf()
{
  NonterminalSymbol result = start_symbol;
  if (isSelf()) return result;
  removeRedundantSymbols();
  result = removeERules();
  // grammar should be without e-rules now
  removeSimpleRules();
  // grammar without e-rules and simple rules cannot have a cycle
  // the last procedure can generate inaccessible (and thus redundant) symbols,
  // so we must remove them
  removeRedundantSymbols();
  return result;
}

bool
ContextfreeGrammar::isinChomskyNormalForm()
{
  // grammar must be without e-rules
  if (!(isERulesfree())) return false;
  // every rule must have one of these forms: A->BC, A->a
  vector<Symbol> str;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) {
    str = iter_r->right;
    if ((str.size() != 0) && (str.size() != 1) && (str.size() != 2))
      return false;
    if (str.size() == 1) if (!isSymbolTerminal(str.front())) return false;
    if (str.size() == 2)
      if (!((isSymbolNonterminal(str.front())) &&
	    (isSymbolNonterminal(str[1]))))
	return false;
  }
  return true;
}

map<NonterminalSymbol, vector<Symbol> > 
ContextfreeGrammar::toChomskyNormalForm()
{
  map<NonterminalSymbol, vector<Symbol> > result_map;
  if (isinChomskyNormalForm()) return result_map;
  // grammar must be self and without simple rules
  if (!isSelf()) { setError(E_GC_NOT_SELF); return result_map; }
  if (!getSimpleRules().empty()) { 
    setError(E_GC_HAS_SIMPLE_RULES); return result_map;
  }
  NonterminalSymbol new_nt;
  vector<Symbol> new_nt_v;
  Rule new_rule;
  set<Rule> rules_cp = rules;
  // cut rules like A->X_1...X_k, k > 2
  for (set<Rule>::const_iterator iter_r = rules_cp.begin();
       iter_r != rules_cp.end(); ++iter_r)
    if (iter_r->right.size() > 2) {
      rules.erase(*iter_r);
      new_rule.left = iter_r->left;
      new_rule.right = vector<Symbol>(1, iter_r->right.front());
      new_nt_v = iter_r->right;
      while (new_nt_v.size() > 2) {
	new_nt_v.erase(new_nt_v.begin());
	new_nt = 0;
	for (map<NonterminalSymbol, vector<Symbol> >::const_iterator
	       iter_map = result_map.begin(); iter_map != result_map.end();
	     ++iter_map)
	  if (iter_map->second == new_nt_v) { 
	    new_nt = iter_map->first; break;
	  }
	if (!new_nt) {
	  new_nt = addNonterminalSymbol(); result_map[new_nt] = new_nt_v;
	}
	new_rule.right.push_back(new_nt);
	rules.insert(new_rule);
	new_rule.left = vector<Symbol>(1, new_nt);
	new_rule.right = vector<Symbol>(1, new_nt_v.front());
      }
      new_rule.right = new_nt_v;
      rules.insert(new_rule);
    }
  rules_cp = rules;
  // replace terminals for nonterminals in rules like A->XY
  for (set<Rule>::const_iterator iter_r = rules_cp.begin();
       iter_r != rules_cp.end(); ++iter_r) {
    new_rule.right.clear();
    if (iter_r->right.size() == 2) {
      rules.erase(*iter_r);
      for (vector<Symbol>::const_iterator iter_sy = iter_r->right.begin();
	   iter_sy != iter_r->right.end(); ++iter_sy)
	if (isSymbolTerminal(*iter_sy)) {
	  new_nt = 0;
	  for (map<NonterminalSymbol, vector<Symbol> >::const_iterator
		 iter_map = result_map.begin(); iter_map != result_map.end();
	       ++iter_map)
	    if ((iter_map->second.size() == 1) &&
		(iter_map->second.front() == *iter_sy)) { 
	      new_nt = iter_map->first; break;
	    }
	  if (!new_nt) {
	    new_nt = addNonterminalSymbol(); 
	    result_map[new_nt] = vector<Symbol>(1, *iter_sy);
	  }
	  new_rule.right.push_back(new_nt);
	  // add new rule like a'->a
	  Rule rl; rl.left = vector<Symbol>(1, new_nt); 
	  rl.right = vector<Symbol>(1, *iter_sy); rules.insert(rl);
	}
	else new_rule.right.push_back(*iter_sy);
      new_rule.left = iter_r->left;
      rules.insert(new_rule);
    }
  }
  return result_map;
}

set<NonterminalSymbol>
ContextfreeGrammar::getRecursiveSymbols()
{
  set<NonterminalSymbol> result;
  set<NonterminalSymbol> rs;
  set<NonterminalSymbol> test_set;
  set<Symbol>::iterator iter_sy_f;
  set<Rule> test_rules, rules_;
  bool inserted;
  for (set<NonterminalSymbol>::const_iterator
	 iter_sy = nonterminal_symbols.begin();
       iter_sy != nonterminal_symbols.end(); ++iter_sy) {
    rs.clear(); rs.insert(*iter_sy);
    test_set.clear(); test_set.insert(*iter_sy);
    test_rules = rules;
    do {
      inserted = false; 
      rules_ = test_rules;
      for (set<Rule>::const_iterator iter_r = rules_.begin();
	   iter_r != rules_.end(); ++iter_r) {
	iter_sy_f = test_set.find(iter_r->left.front());
	if (iter_sy_f != test_set.end()) {
	  for (vector<Symbol>::const_iterator iter_sy1 = iter_r->right.begin();
	       iter_sy1 != iter_r->right.end(); ++iter_sy1) {
	    if (isSymbolNonterminal(*iter_sy1)) {
	      pair<set<Symbol>::iterator, bool> p = rs.insert(*iter_sy1);
	      if (p.second) inserted = true;
	      if (*iter_sy1 == *iter_sy)
		// *iter_sy1 is recursive
		result.insert(*iter_sy1);
	    }
	  }
	  test_rules.erase(*iter_r);
	}
      }
      test_set.insert(rs.begin(), rs.end());
    } while (inserted);
  }
  return result;
}

bool
ContextfreeGrammar::isRecursive()
{
  // grammar is recursive if has at least one recursive nonterminal symbol
  if (!getRecursiveSymbols().empty()) return true;
  return false;
}

map<NonterminalSymbol, NonterminalSymbol>
ContextfreeGrammar::eliminateLeftRecursion()
{
  map<NonterminalSymbol, NonterminalSymbol> result_map;
  // grammar must be self
  if (!isSelf()) { setError(E_GC_NOT_SELF); return result_map; }
  // we need sorted nonterminal symbols
  vector<NonterminalSymbol> nts_v;
  for (set<NonterminalSymbol>::const_iterator 
	 iter_sy = nonterminal_symbols.begin();
       iter_sy != nonterminal_symbols.end(); ++iter_sy)
    nts_v.push_back(*iter_sy);
  if (nts_v.empty()) return result_map;
  NonterminalSymbol new_nt;
  Rule new_rule;
  set<Rule> lr_rules, rest_rules;
  unsigned int i = 0, j;
  while (1) {
    lr_rules.clear(); rest_rules.clear();
    // find rules with straight left recursion
    for (set<Rule>::const_iterator iter_r = rules.begin();
	 iter_r != rules.end(); ++iter_r)
      if (iter_r->left.front() == nts_v[i]) {
	if ((!iter_r->right.empty()) && (iter_r->right.front() == nts_v[i]))
	  lr_rules.insert(*iter_r);
	else rest_rules.insert(*iter_r);
      }
    // eliminate straight left recursion
    if (!lr_rules.empty()) {
      if (!rest_rules.empty()) {
	new_nt = addNonterminalSymbol(); result_map[new_nt] = nts_v[i];
      }
      // if there's no rest rule, simply convert left recursive rule A->Aab
      // to right recursive A->abA
      else new_nt = nts_v[i];
      for (set<Rule>::const_iterator iter_r = rest_rules.begin();
	   iter_r != rest_rules.end(); ++iter_r) {
	new_rule = *iter_r; new_rule.right.push_back(new_nt);
	rules.insert(new_rule);
      }
      for (set<Rule>::const_iterator iter_r = lr_rules.begin();
	   iter_r != lr_rules.end(); ++iter_r) {
	rules.erase(*iter_r);
	new_rule = *iter_r; new_rule.left[0] = new_nt;
	new_rule.right.erase(new_rule.right.begin());
	rules.insert(new_rule);
	new_rule.right.push_back(new_nt); rules.insert(new_rule);
      }
    }
    if (i == nts_v.size() - 1) break; else { ++i; j = 0; }
    while (1) {
      rest_rules.clear();
      // eliminate non-straight left recursion
      for (set<Rule>::const_iterator iter_r = rules.begin();
	   iter_r != rules.end(); ++iter_r)
	if (iter_r->left.front() == nts_v[j]) rest_rules.insert(*iter_r);
      lr_rules = rules;
      for (set<Rule>::const_iterator iter_r = lr_rules.begin();
	   iter_r != lr_rules.end(); ++iter_r)
	if ((!iter_r->right.empty()) && (iter_r->left.front() == nts_v[i]) && 
	    (iter_r->right.front() == nts_v[j])) {
	  rules.erase(*iter_r);
	  for (set<Rule>::const_iterator iter_r1 = rest_rules.begin();
	       iter_r1 != rest_rules.end(); ++iter_r1) {
	    new_rule = *iter_r; new_rule.right.erase(new_rule.right.begin());
	    new_rule.right.insert
	      (new_rule.right.begin(), iter_r1->right.begin(),
	       iter_r1->right.end());
	    rules.insert(new_rule);
	  }
	}
      if (j == i - 1) break; else ++j;
    }
  }
  return result_map;
}

bool
ContextfreeGrammar::isinGreibachNormalForm()
{
  // grammar must be without e-rules
  if (!(isERulesfree())) return false;
  // every rule must have this form: A->aXYZ...
  vector<Symbol> str;
  vector<Symbol>::const_iterator iter_sy;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) {
    str = iter_r->right;
    if (str.size() == 0) continue; // S->e
    if (!isSymbolTerminal(str.front())) return false;
    iter_sy = str.begin(); ++iter_sy;
    for (; iter_sy != str.end(); ++iter_sy) 
      if (!isSymbolNonterminal(*iter_sy)) return false;
  }
  return true;
}

map<NonterminalSymbol, TerminalSymbol>
ContextfreeGrammar::toGreibachNormalForm()
{
  map<NonterminalSymbol, TerminalSymbol> result_map;
  if (isinGreibachNormalForm()) return result_map;
  // grammar must be self and left recursion free
  if (!isSelf()) { setError(E_GC_NOT_SELF); return result_map; }
  if (isLeftRecursive()) { 
    setError(E_GC_IS_LEFT_RECURSIVE); return result_map;
  }
  // we need specially sorted nonterminal symbols
  vector<NonterminalSymbol> nts_v;
  for (set<NonterminalSymbol>::const_iterator 
	 iter_sy = nonterminal_symbols.begin();
       iter_sy != nonterminal_symbols.end(); ++iter_sy)
    nts_v.push_back(*iter_sy);
  // really nice, nice hack :)
  // STL sort function (uses introsort alg.) sorts badly when there's no
  // way how to compare some items (and thus they are considered equal), so
  // I use partial_sort (uses heapsort alg.) which works fine
  // sort(nts_v.begin(), nts_v.end(), *this);
  partial_sort(nts_v.begin(), nts_v.end(), nts_v.end(), *this);
  NonterminalSymbol new_nt;
  Rule new_rule;
  set<Rule> i_rules, j_rules;
  unsigned int i;
  if (nts_v.size() > 1) {
    i = nts_v.size() - 1;
    // remove rules with right part starting with nonterminal symbol
    do {
      --i;
      for (unsigned int j = nts_v.size() - 1; j > i; --j) {
	j_rules.clear();
	for (set<Rule>::const_iterator iter_r = rules.begin();
	     iter_r != rules.end(); ++iter_r)
	  if (iter_r->left.front() == nts_v[j]) j_rules.insert(*iter_r);
	i_rules = rules;
	for (set<Rule>::const_iterator iter_r = i_rules.begin();
	     iter_r != i_rules.end(); ++iter_r)
	  if ((iter_r->left.front() == nts_v[i]) && (!iter_r->right.empty()) &&
	      (iter_r->right.front() == nts_v[j])) {
	    rules.erase(*iter_r);
	    for (set<Rule>::const_iterator iter_r1 = j_rules.begin();
		 iter_r1 != j_rules.end(); ++iter_r1) {
	      new_rule = *iter_r; new_rule.right.erase(new_rule.right.begin());
	      new_rule.right.insert
		(new_rule.right.begin(), iter_r1->right.begin(),
		 iter_r1->right.end());
	      rules.insert(new_rule);
	    }
	  }
      }
    } while (i > 0);
  }
  i_rules = rules;
  for (set<Rule>::const_iterator iter_r = i_rules.begin();
       iter_r != i_rules.end(); ++iter_r) {
    new_rule.right.clear();
    if (!iter_r->right.empty()) { // S->e
      rules.erase(*iter_r);
      // first terminal symbol stays
      vector<Symbol>::const_iterator iter_sy = iter_r->right.begin();
      new_rule.right.push_back(*iter_sy); ++iter_sy;
      for (; iter_sy != iter_r->right.end(); ++iter_sy)
	if (isSymbolTerminal(*iter_sy)) {
	  new_nt = 0;
	  for (map<NonterminalSymbol, TerminalSymbol>::const_iterator
		 iter_map = result_map.begin(); iter_map != result_map.end();
	       ++iter_map)
	    if (iter_map->second == *iter_sy) {
	      new_nt = iter_map->first; break;
	    }
	  if (!new_nt) {
	    new_nt = addNonterminalSymbol(); 
	    result_map[new_nt] = *iter_sy;
	  }
	  new_rule.right.push_back(new_nt);
	  // add new rule like a'->a
	  Rule rl; rl.left = vector<Symbol>(1, new_nt); 
	  rl.right = vector<Symbol>(1, *iter_sy); rules.insert(rl);
	}
	else new_rule.right.push_back(*iter_sy);
      new_rule.left = iter_r->left;
      rules.insert(new_rule);
    }
  }
  return result_map;
}

bool
ContextfreeGrammar::isLeftRecursive()
{
  set<NonterminalSymbol> rs;
  set<NonterminalSymbol> test_set;
  set<Symbol>::iterator iter_sy_f;
  set<Rule> test_rules, rules_;
  bool inserted;
  for (set<NonterminalSymbol>::const_iterator
	 iter_sy = nonterminal_symbols.begin();
       iter_sy != nonterminal_symbols.end(); ++iter_sy) {
    rs.clear(); rs.insert(*iter_sy);
    test_set.clear(); test_set.insert(*iter_sy);
    test_rules = rules;
    do {
      inserted = false; 
      rules_ = test_rules;
      for (set<Rule>::const_iterator iter_r = rules_.begin();
	   iter_r != rules_.end(); ++iter_r) {
	iter_sy_f = test_set.find(iter_r->left.front());
	if ((iter_sy_f != test_set.end()) && (!iter_r->right.empty()) &&
	    (isSymbolNonterminal(iter_r->right.front()))) {
	  pair<set<Symbol>::iterator, bool> 
	    p = rs.insert(iter_r->right.front());
	  if (p.second) inserted = true;
	  if (iter_r->right.front() == *iter_sy)
	    // iter_r->right.front() is left recursive
	    return true;
	  test_rules.erase(*iter_r);
	}
      }
      test_set.insert(rs.begin(), rs.end());
    } while (inserted);
  }
  return false;
}

bool
ContextfreeGrammar::operator()(NonterminalSymbol nt1, NonterminalSymbol nt2)
{
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r)
    if ((!iter_r->right.empty()) && (iter_r->left.front() == nt1) &&
	(iter_r->right.front() == nt2))
      return true;
  return false;
}
