/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "error.h"
#include "finite_automata.h"

FiniteAutomata::FiniteAutomata()
{

}

FiniteAutomata::~FiniteAutomata()
{

}

bool
FiniteAutomata::addTransition(Transition& transition, State from_state,
			      EntrySymbol symbol, State to_state)
{
  // check if both states are in the states set
  set<State>::iterator iter_st = states.find(from_state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  iter_st = states.find(to_state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  // check if the symbol is in the entry_symbols set
  // CHOICE: if not, add it - now NO
  set<EntrySymbol>::iterator iter_sy = entry_symbols.find(symbol);
  if (iter_sy == entry_symbols.end()) { 
    setError(E_A_ENTRY_SYMBOL_BAD); return false;
  }
  transition.from_state = from_state;
  transition.to_state = to_state;
  transition.entry_symbol = symbol;
  transitions.insert(transition);
  return true;
}

bool
FiniteAutomata::addTransition(State from_state, EntrySymbol symbol,
			      State to_state)
{
  Transition tran;
  return addTransition(tran, from_state, symbol, to_state);
}

bool
FiniteAutomata::removeTransition(State from_state, EntrySymbol symbol,
				 State to_state)
{
  // check if both states are in the states set
  set<State>::iterator iter_st = states.find(from_state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  iter_st = states.find(to_state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  // check if the symbol is in the entry_symbols set
  set<EntrySymbol>::iterator iter_sy = entry_symbols.find(symbol);
  if (iter_sy == entry_symbols.end()) {
    setError(E_A_ENTRY_SYMBOL_BAD); return false;
  }
  Transition transition;
  transition.from_state = from_state;
  transition.to_state = to_state;
  transition.entry_symbol = symbol;
  return Automata::removeTransition(transition);
}

bool
FiniteAutomata::setConfiguration(State state, vector<EntrySymbol> sentence)
{
  // check if the state is in the states set
  set<State>::iterator iter_st = states.find(state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  // check if all symbols in the sentence are in the entry_symbols set
  // CHOICE: if not, add it - now NO
  set<EntrySymbol>::iterator iter_sy_f;
  for (vector<EntrySymbol>::const_iterator iter_sy = sentence.begin();
       iter_sy != sentence.end(); ++iter_sy) {
    iter_sy_f = entry_symbols.find(*iter_sy);
    if (iter_sy_f == entry_symbols.end()) { 
      setError(E_A_ENTRY_SYMBOL_BAD); return false;
    }
  }
  c_state = state;
  c_sentence = sentence;
  return true;
}

inline void
FiniteAutomata::getConfiguration(State& state, vector<EntrySymbol>& sentence)
{
  state = c_state;
  sentence = c_sentence;
}

set<State>
FiniteAutomata::getTransitionDestinations()
{
  set<State> result;
  if (c_sentence.empty()) return result;
  return _getTransitionDestinations(c_state, c_sentence.front());
}

set<State>
FiniteAutomata::_getTransitionDestinations(State state, EntrySymbol symbol)
{
  set<State> result;
  if (state == State_NULL) return result;
  // we want to_state of transitions which have from_state == state and
  // entry_symbol == symbol
  for (set<Transition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t)
    if ((iter_t->from_state == state) &&
	(iter_t->entry_symbol == symbol))
      result.insert(iter_t->to_state);
  return result;
}

inline bool
FiniteAutomata::doTransition(State state)
{
  return _doTransition(c_state, state, c_sentence);
}

bool
FiniteAutomata::_doTransition(State& from_state, State to_state,
			      vector<EntrySymbol>& sentence)
{
  set<State> dest_states;
  if (sentence.empty()) { setError(E_AF_SENTENCE_EMPTY); return false; }
  dest_states = _getTransitionDestinations(from_state, sentence.front());
  // check if the to_state is among dest_states
  set<State>::iterator iter_st = dest_states.find(to_state);
  if (iter_st == dest_states.end()) { 
    setError(E_AF_STATE_CANT_REACH); return false;
  }
  // move to the to_state
  from_state = to_state;
  // erase the first symbol of the sentence
  sentence.erase(sentence.begin());
  return true;
}

inline bool
FiniteAutomata::checkSentence(vector<EntrySymbol> sentence)
{
  return _checkSentence(start_state, sentence);
}

bool
FiniteAutomata::_checkSentence(State state, vector<EntrySymbol> sentence)
{
  set<State> dest_states;
  if (!sentence.empty()) 
    dest_states = _getTransitionDestinations(state, sentence.front());
  State new_state;
  vector<EntrySymbol> rest_of_sentence = sentence;
  for (set<State>::const_iterator iter_st = dest_states.begin(); 
       iter_st != dest_states.end(); ++iter_st) {
    // we must have the same conditions for each *iter state
    new_state = state;
    rest_of_sentence = sentence;
    _doTransition(new_state, *iter_st, rest_of_sentence);
    // check the rest of the sentence
    if (_checkSentence(new_state, rest_of_sentence)) return true;
    if (rest_of_sentence.empty()) {
      set<State>::iterator iter_st = end_states.find(new_state);
      if (iter_st != end_states.end()) return true;
    }
  }
  // check if all entry symbols were accepted
  if (rest_of_sentence.empty()) {
    // check if the state is end state
    set<State>::iterator iter_st = end_states.find(new_state);
    if (iter_st != end_states.end()) return true;
  }
  return false;
}

bool
FiniteAutomata::isDeterministic()
{
  for (set<Transition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t)
    if (_getTransitionDestinations
	(iter_t->from_state, iter_t->entry_symbol).size() > 1) return false;
  return true;
}

map<State, set<State> >
FiniteAutomata::toDeterministic()
{
  map<State, set<State> > result_map;
  if (isDeterministic()) return result_map;
  FiniteAutomata new_automata;
  State new_state, state_act;
  set<State> new_set_state, set_state_act, set_state_1;
  bool end_state;
  map<State, bool> done_states;
  // entry symbols stay the same
  new_automata.entry_symbols = entry_symbols;
  // add start state
  state_act = new_automata.addState();
  new_automata.start_state = state_act;
  set_state_act.insert(start_state);
  // map it
  result_map[state_act] = set_state_act;
  // this state is not done
  done_states[state_act] = false;
  bool non_done_left = true;
  // for all non-done new states
  while (non_done_left) {
    end_state = false;
    // for all entry symbols
    for (set<EntrySymbol>::const_iterator iter_sy = entry_symbols.begin();
	 iter_sy != entry_symbols.end(); ++iter_sy) {
      new_set_state.clear();
      // for all orig. states from actual non-done state (set_state_act)
      for (set<State>::const_iterator iter_st = set_state_act.begin();
	   iter_st != set_state_act.end(); ++iter_st) {
	// make dest. set state as a union of orig. dest. states
	set_state_1 = _getTransitionDestinations(*iter_st, *iter_sy);
	new_set_state.insert(set_state_1.begin(), set_state_1.end());
	// check if the orig. state (*iter_st) is end state; if it is, actual 
	// state will be too
	if (!end_state) {
	  set<State>::iterator iter_st_f = end_states.find(*iter_st);
	  if (iter_st_f != end_states.end()) end_state = true;
	}
      }
      // no dest. set state -> next entry symbol
      if (new_set_state.empty()) continue;
      // add new set state (if it isn't already in the automata states set)
      map<State, set<State> >::const_iterator iter_map = result_map.begin();
      for (; iter_map != result_map.end(); ++iter_map) 
	if (iter_map->second == new_set_state) break;
      if (iter_map == result_map.end()) {
	new_state = new_automata.addState();
	result_map[new_state] = new_set_state;
	done_states[new_state] = false;
      }
      else new_state = iter_map->first;
      // add new transition
      new_automata.addTransition(state_act, *iter_sy, new_state);
    }
    // set end state
    if (end_state) new_automata.end_states.insert(state_act);
    // actual state is done
    done_states[state_act] = true;
    // set new actual non-done state
    map<State, bool>::const_iterator iter_map = done_states.begin();
    for (; iter_map != done_states.end(); ++iter_map) 
      if (!iter_map->second) break;
    if (iter_map != done_states.end()) {
      state_act = iter_map->first; set_state_act = result_map[state_act];
    }
    else non_done_left = false;
  }
  // set configuration
  new_automata.c_state = new_automata.start_state;
  new_automata.c_sentence = c_sentence;
  // new_automata will be this automata
  *this = new_automata;
  // return states map
  return result_map;
}

set<State>
FiniteAutomata::getInaccessibleStates()
{
  set<State> accessible_states;
  set<State> states_1, states_2;
  set<State> *iter_states, *insert_states;
  // start state is accessible
  states_1.insert(start_state);
  accessible_states.insert(start_state);
  iter_states = &states_1; insert_states = &states_2;
  while (!iter_states->empty()) {
    for (set<State>::const_iterator iter_st = iter_states->begin();
	 iter_st != iter_states->end(); ++iter_st)
      for (set<Transition>::const_iterator iter_t = transitions.begin();
	   iter_t != transitions.end();
	   ++iter_t)
	if (iter_t->from_state == *iter_st) {
	  State new_state = iter_t->to_state;
	  // check if to_state isn't already among accessible states
	  // if yes, don't iter through it again
	  set<State>::iterator iter_st_f = accessible_states.find(new_state);
	  if (iter_st_f == accessible_states.end()) {
	    insert_states->insert(new_state);
	    accessible_states.insert(new_state);
	  }
	}
    // new iter_states = insert_states
    set<State> *p_states = iter_states;
    iter_states = insert_states; insert_states = p_states;
    insert_states->clear();
  }
  set<State> result;
  set_difference(states.begin(), states.end(), accessible_states.begin(),
		 accessible_states.end(), insert_iterator<set<State> >
		 (result, result.begin()));
  return result;
}

map<State, set<State> >
FiniteAutomata::toReduced()
{
  map<State, set<State> > result_map;
  // reduced automata can be constructed only to deterministic automata
  if (!isDeterministic()) { 
    setError(E_AF_NOT_DETERMINISTIC); return result_map;
  }
  // eliminate inaccessible states
  set<State> i_states = getInaccessibleStates();
  for (set<State>::const_iterator iter_st = i_states.begin(); 
       iter_st != i_states.end(); ++iter_st) removeState(*iter_st);
  FiniteAutomata new_automata;
  // entry symbols stay the same
  new_automata.entry_symbols = entry_symbols;
  set<set<State> > eq_classes = getEquivalentionRelation();
  // create new states and states map
  for (set<set<State> >::const_iterator iter_set = eq_classes.begin(); 
	 iter_set != eq_classes.end(); ++iter_set)
    result_map[new_automata.addState()] = *iter_set;
  set<State> set_state;
  // create new transitions
  for (set<Transition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end();
       ++iter_t) {
    State from_state = State_NULL, to_state = State_NULL;
    for (map<State, set<State> >::const_iterator
	   iter_map = result_map.begin(); iter_map != result_map.end();
	 ++iter_map) {
      set_state = iter_map->second;
      set<State>::iterator iter_st_f = set_state.find(iter_t->from_state);
      if (iter_st_f != set_state.end()) from_state = iter_map->first;
      iter_st_f = set_state.find(iter_t->to_state);
      if (iter_st_f != set_state.end()) to_state = iter_map->first;
    }
    new_automata.addTransition(from_state, iter_t->entry_symbol, to_state);
  }
  for (map<State, set<State> >::const_iterator
	 iter_map = result_map.begin(); iter_map != result_map.end();
       ++iter_map) {
    set_state = iter_map->second;
    // end states
    set<State> end_sts;
    set_intersection(set_state.begin(), set_state.end(), end_states.begin(),
		     end_states.end(), insert_iterator<set<State> >
		     (end_sts, end_sts.begin()));
    if (!end_sts.empty()) new_automata.end_states.insert(iter_map->first);
    // start state
    set<State>::iterator iter_st_f = set_state.find(start_state);
    if (iter_st_f != set_state.end()) 
      new_automata.start_state = iter_map->first;
    // config. state
    iter_st_f = set_state.find(c_state);
    if (iter_st_f != set_state.end()) 
      new_automata.c_state = iter_map->first;
  }
  new_automata.c_sentence = c_sentence;
  // new_automata will be this automata
  *this = new_automata;
  // return states map
  return result_map;
}

set<set<State> >
FiniteAutomata::getEquivalentionRelation()
{
  set<set<State> > sets, sets_prev;
  // create eq. relation 0 = 2 classes: states which either are end or are not
  set<State> non_end_states;
  set_difference(states.begin(), states.end(), end_states.begin(),
		 end_states.end(), insert_iterator<set<State> >
		 (non_end_states, non_end_states.begin()));
  sets.insert(end_states);
  sets.insert(non_end_states);
  set<State> split_set;
  do {
    // create new eq. relation (one higher)
    sets_prev = sets;
    sets.clear();
    // for all classes in relation
    for (set<set<State> >::const_iterator iter_set = sets_prev.begin(); 
	 iter_set != sets_prev.end(); ++iter_set) {
      split_set = *iter_set;
      // split class to eq. states
      while (!split_set.empty()) {
	set<State> new_set; new_set.insert(*(split_set.begin()));
	for (set<State>::const_iterator iter_st = ++(split_set.begin()); 
	     iter_st != split_set.end(); ++iter_st) {
	  bool together = true;
	  // iterate throught all entry symbols
	  for (set<EntrySymbol>::const_iterator 
		 iter_sy = entry_symbols.begin(); 
	       iter_sy != entry_symbols.end(); ++iter_sy) {
	    set<State> dests1 = 
	      _getTransitionDestinations(*(split_set.begin()), *iter_sy);
	    if (dests1.empty()) continue;
	    set<State> dests2 = _getTransitionDestinations(*iter_st, *iter_sy);
	    if (dests2.empty()) continue;
	    if (!areStatesinOneSet(*(dests1.begin()), *(dests2.begin()),
				   sets_prev))
	      { together = false; break; }
	  }
	  // if dest. states are equivalent, they will be in one new class
	  if (together) new_set.insert(*iter_st);
	}
	sets.insert(new_set);
	set<State> split_set_rest;
	set_difference(split_set.begin(), split_set.end(), new_set.begin(),
		       new_set.end(), insert_iterator<set<State> >
		       (split_set_rest, split_set_rest.begin()));
	split_set = split_set_rest;
      }
    }
  } while (sets != sets_prev);
  return sets;
}

bool
FiniteAutomata::areStatesinOneSet(State st1, State st2, set<set<State> > sets)
{
  bool found1, found2;
  set<State>::iterator iter_st;
  for (set<set<State> >::const_iterator iter_set = sets.begin(); 
       iter_set != sets.end(); ++iter_set) {
    iter_st = iter_set->find(st1);
    if (iter_st != iter_set->end()) found1 = true; else found1 = false;
    iter_st = iter_set->find(st2);
    if (iter_st != iter_set->end()) found2 = true; else found2 = false;
    if ((found1) && (found2)) return true;
    if ((found1) || (found2)) return false;
  }
  // can't be reached, each state must be in some set
  return true;
}
