/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "finite_automata_l.h"

const char string_space = ' ';
const char string_lbracket = '(';
const char string_rbracket = ')';

FiniteAutomata_l::FiniteAutomata_l()
{
  automata = new FiniteAutomata;
  Automata_l::automata = automata;
}

FiniteAutomata_l::~FiniteAutomata_l()
{
  delete automata;
  automata = NULL;
  Automata_l::automata = NULL;
}

inline bool
FiniteAutomata_l::addTransition(string from_state_id, string symbol_id,
				string to_state_id)
{
  return automata->addTransition
    (string_to_State(from_state_id), string_to_EntrySymbol(symbol_id),
     string_to_State(to_state_id));
}

inline bool
FiniteAutomata_l::removeTransition(string from_state_id, string symbol_id,
				   string to_state_id)
{
  return automata->removeTransition
    (string_to_State(from_state_id), string_to_EntrySymbol(symbol_id),
     string_to_State(to_state_id));
}

bool
FiniteAutomata_l::setConfiguration(string state_id, string sentence)
{
  vector<EntrySymbol> sent;
  string symbol;
  // parse sentence string
  for (string::const_iterator iter_s = sentence.begin();
       iter_s != sentence.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      sent.push_back(string_to_EntrySymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) sent.push_back(string_to_EntrySymbol(symbol));
  return automata->setConfiguration(string_to_State(state_id), sent);
}

void
FiniteAutomata_l::getConfiguration(string& state, string& sentence)
{
  State st;
  vector<EntrySymbol> sent;
  automata->getConfiguration(st, sent);
  // return st and sent looks
  state = states[st].getLook(st, getStateFlags(st));
  sentence = string_lbracket; sentence += "entry_symbols";
  for (vector<EntrySymbol>::const_iterator iter_sy = sent.begin();
       iter_sy != sent.end(); ++iter_sy)
    sentence += string_space + entry_symbols[*iter_sy].getLook(*iter_sy);
  sentence += string_rbracket;
}

string
FiniteAutomata_l::getLook()
{
  string result;
  result += string_lbracket; result += "finite_automata";
  result += string_space;
  // State looks
  result += string_lbracket; result += "states";
  set<State> sts = automata->getStates();
  for (set<State>::const_iterator iter_st = sts.begin(); iter_st != sts.end();
       ++iter_st)
    result += string_space +
      states[*iter_st].getLook(*iter_st, getStateFlags(*iter_st));
  result += string_rbracket; result += string_space;
  // EntrySymbol looks
  result += string_lbracket; result += "entry_symbols";
  set<EntrySymbol> symbols = automata->getEntrySymbols();
  for (set<EntrySymbol>::const_iterator iter_sy = symbols.begin();
       iter_sy != symbols.end(); ++iter_sy)
    result += string_space + entry_symbols[*iter_sy].getLook(*iter_sy);
  result += string_rbracket; result += string_space;
  // transition looks
  // transition look: (transition 
  // _from_state_look_ _entry_symbol_look_ _to_state_look_)
  result += string_lbracket; result += "transitions";
  set<Transition> trans = automata->getTransitions();
  for (set<Transition>::const_iterator iter_t = trans.begin();
       iter_t != trans.end(); ++iter_t) {
    result += string_space; result += string_lbracket; result += "transition";
    result += string_space + states[iter_t->from_state].getLook
      (iter_t->from_state, getStateFlags(iter_t->from_state));
    result += string_space + entry_symbols[iter_t->entry_symbol].getLook
      (iter_t->entry_symbol);
    result += string_space + states[iter_t->to_state].getLook
      (iter_t->to_state, getStateFlags(iter_t->to_state));
    result += string_rbracket;
  }
  result += string_rbracket; result += string_rbracket;
  return result;
}

string
FiniteAutomata_l::getTransitionDestinations()
{
  string result;
  set<State> dests = automata->getTransitionDestinations();
  // return dests look
  result = string_lbracket; result += "states";
  for (set<State>::const_iterator iter_st = dests.begin();
       iter_st != dests.end(); ++iter_st)
    result += string_space + 
      states[*iter_st].getLook(*iter_st, getStateFlags(*iter_st));
  result += string_rbracket;
  return result;
}

inline bool
FiniteAutomata_l::doTransition(string state_id)
{
  return automata->doTransition(string_to_State(state_id));
}

bool
FiniteAutomata_l::checkSentence(string sentence)
{
  vector<EntrySymbol> sent;
  string symbol;
  // parse sentence string
  for (string::const_iterator iter_s = sentence.begin();
       iter_s != sentence.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      sent.push_back(string_to_EntrySymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) sent.push_back(string_to_EntrySymbol(symbol));
  return automata->checkSentence(sent);
}

inline bool
FiniteAutomata_l::isDeterministic()
{
  return automata->isDeterministic();
}

void
FiniteAutomata_l::toDeterministic()
{
  map<State, set<State> > state_map = automata->toDeterministic();
  if (state_map.empty()) return;
  // build new states (state map) and new looks
  map<State, State_l> new_states;
  set<State> sts;
  string name;
  for (map<State, set<State> >::const_iterator iter_map = state_map.begin();
       iter_map != state_map.end(); ++iter_map) {
    // build states names from old states names
    sts = iter_map->second;
    name = "[";
    for (set<State>::const_iterator iter_st = sts.begin();
	 iter_st != sts.end(); ++iter_st)
      name += states[*iter_st].getName() + ",";
    string::iterator last = name.end(); --last;
    name.erase(last);
    name += "]";
    new_states[iter_map->first] = State_l(name);
  }
  states = new_states;
}

string
FiniteAutomata_l::getInaccessibleStates()
{
  string result;
  set<State> i_states = automata->getInaccessibleStates();
  // return states look
  result = string_lbracket; result += "states";
  for (set<State>::const_iterator iter_st = i_states.begin(); 
       iter_st != i_states.end(); ++iter_st)
    result += string_space + 
      states[*iter_st].getLook(*iter_st, getStateFlags(*iter_st));
  result += string_rbracket;
  return result;
}

void
FiniteAutomata_l::toReduced()
{
  map<State, set<State> > state_map = automata->toReduced();
  if (state_map.empty()) return;
  // build new states (state map) and new looks
  map<State, State_l> new_states;
  set<State> sts;
  string name;
  for (map<State, set<State> >::const_iterator iter_map = state_map.begin();
       iter_map != state_map.end(); ++iter_map) {
    // build states names from old states names
    sts = iter_map->second;
    name = "[";
    for (set<State>::const_iterator iter_st = sts.begin();
	 iter_st != sts.end(); ++iter_st)
      name += states[*iter_st].getName() + ",";
    string::iterator last = name.end(); --last;
    name.erase(last);
    name += "]";
    new_states[iter_map->first] = State_l(name);
  }
  states = new_states;
}

const FiniteAutomata_l &
FiniteAutomata_l::operator=(const FiniteAutomata_l &a)
{
  if (this != &a) {
    states = a.states;
    entry_symbols = a.entry_symbols;
    *automata = *(a.automata);
    *(Automata_l::automata) = *(a.automata);
  }
  return *this;
}
