/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "error.h"
#include "grammar.h"

Grammar::Grammar()
{
  start_symbol = NonterminalSymbol_NULL;
};

Grammar::~Grammar()
{

};

NonterminalSymbol
Grammar::addNonterminalSymbol()
{
  NonterminalSymbol new_symbol = NonterminalSymbol_MIN;
  for (set<NonterminalSymbol>::const_iterator 
	 iter_sy = nonterminal_symbols.begin();
       iter_sy != nonterminal_symbols.end(); ++iter_sy, ++new_symbol)
    if (*iter_sy != new_symbol) break;
  if (new_symbol > NonterminalSymbol_MAX) { 
    setError(E_G_NONTERMINAL_SYMBOL_MAX); return NonterminalSymbol_NULL;
  }
  nonterminal_symbols.insert(new_symbol);
  // if there's only one nonterminal symbol it's start
  if (nonterminal_symbols.size() == 1) start_symbol = new_symbol;
  return new_symbol;
}

bool
Grammar::removeNonterminalSymbol(NonterminalSymbol symbol)
{
  set<NonterminalSymbol>::iterator
    iter_sy_f = nonterminal_symbols.find(symbol);
  if (iter_sy_f == nonterminal_symbols.end()) { 
    setError(E_G_NONTERMINAL_SYMBOL_BAD); return false;
  }
  // erase also rules which contain the symbol
  set<Rule> rules = find_rules_with_nonterminal_symbol(symbol);
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) removeRule(*iter_r);
  // if we're erasing start symbol we must set new one
  if (symbol == start_symbol) {
    if (nonterminal_symbols.size() != 1) {
      if (symbol == *nonterminal_symbols.begin())
	start_symbol = *++nonterminal_symbols.begin();
      else start_symbol = *nonterminal_symbols.begin();
    }
    else start_symbol = NonterminalSymbol_NULL;
  }
  nonterminal_symbols.erase(symbol);
  return true;
}

set<Rule>
Grammar::find_rules_with_nonterminal_symbol(NonterminalSymbol symbol)
{
  set<Rule> result;
  vector<Symbol> str;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) {
    str = iter_r->left;
    for (vector<Symbol>::const_iterator iter_sy = str.begin();
         iter_sy != str.end(); ++iter_sy) 
      if (*iter_sy == symbol) { result.insert(*iter_r); break; }
    str = iter_r->right;
    for (vector<Symbol>::const_iterator iter_sy = str.begin();
         iter_sy != str.end(); ++iter_sy) 
      if (*iter_sy == symbol) { result.insert(*iter_r); break; }
  }    
  return result;
}

inline set<NonterminalSymbol>
Grammar::getNonterminalSymbols()
{
  return nonterminal_symbols;
}

TerminalSymbol
Grammar::addTerminalSymbol()
{
  TerminalSymbol new_symbol = TerminalSymbol_MIN;
  for (set<TerminalSymbol>::const_iterator 
	 iter_sy = terminal_symbols.begin();
       iter_sy != terminal_symbols.end(); ++iter_sy, ++new_symbol)
    if (*iter_sy != new_symbol) break;
  if (new_symbol > TerminalSymbol_MAX) {
    setError(E_G_TERMINAL_SYMBOL_MAX); return TerminalSymbol_NULL;
  }
  terminal_symbols.insert(new_symbol);
  return new_symbol;
}

bool
Grammar::removeTerminalSymbol(TerminalSymbol symbol)
{
  set<TerminalSymbol>::iterator iter_sy_f = terminal_symbols.find(symbol);
  if (iter_sy_f == terminal_symbols.end()) { 
    setError(E_G_TERMINAL_SYMBOL_BAD); return false;
  }
  // erase also rules which contain the symbol
  set<Rule> rules = find_rules_with_terminal_symbol(symbol);
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) removeRule(*iter_r);
  terminal_symbols.erase(symbol);
  return true;
}

set<Rule>
Grammar::find_rules_with_terminal_symbol(TerminalSymbol symbol)
{
  set<Rule> result;
  vector<Symbol> str;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) {
    str = iter_r->left;
    for (vector<Symbol>::const_iterator iter_sy = str.begin();
         iter_sy != str.end(); ++iter_sy) 
      if (*iter_sy == symbol) { result.insert(*iter_r); break; }
    str = iter_r->right;
    for (vector<Symbol>::const_iterator iter_sy = str.begin();
         iter_sy != str.end(); ++iter_sy) 
      if (*iter_sy == symbol) { result.insert(*iter_r); break; }
  }    
  return result;
}

inline set<TerminalSymbol>
Grammar::getTerminalSymbols()
{
  return terminal_symbols;
}

bool
Grammar::addRule(Rule &rule, vector<Symbol> left, vector<Symbol> right)
{
  // check if all symbols in both parts are in the [non]terminal_symbols set
  // CHOICE: if not, add it - now NO
  set<Symbol>::iterator iter_sy_f;
  for (vector<Symbol>::const_iterator iter_sy = left.begin();
       iter_sy != left.end(); ++iter_sy)
    if (isSymbolNonterminal(*iter_sy)) {
      iter_sy_f = nonterminal_symbols.find(*iter_sy);
      if (iter_sy_f == nonterminal_symbols.end()) {
	setError(E_G_NONTERMINAL_SYMBOL_BAD); return false;
      }
    } else {
      iter_sy_f = terminal_symbols.find(*iter_sy);
      if (iter_sy_f == terminal_symbols.end()) {
	setError(E_G_TERMINAL_SYMBOL_BAD); return false;
      }
    }
  for (vector<Symbol>::const_iterator iter_sy = right.begin();
       iter_sy != right.end(); ++iter_sy)
    if (isSymbolNonterminal(*iter_sy)) {
      iter_sy_f = nonterminal_symbols.find(*iter_sy);
      if (iter_sy_f == nonterminal_symbols.end()) {
	setError(E_G_NONTERMINAL_SYMBOL_BAD); return false;
      }
    } else {
      iter_sy_f = terminal_symbols.find(*iter_sy);
      if (iter_sy_f == terminal_symbols.end()) {
	setError(E_G_TERMINAL_SYMBOL_BAD); return false;
      }
    }
  rule.left = left;
  rule.right = right;
  rules.insert(rule);
  return true;
}

bool
Grammar::addRule(vector<Symbol> left, vector<Symbol> right)
{
  Rule rule;
  return addRule(rule, left, right);
}

bool
Grammar::removeRule(Rule rule)
{
  set<Rule>::iterator iter_r_f = rules.find(rule);
  if (iter_r_f == rules.end()) { 
    setError(E_G_RULE_BAD); return false;
  }
  rules.erase(rule);
  return true;
}

bool
Grammar::removeRule(vector<Symbol> left, vector<Symbol> right)
{
  // check if all symbols in both parts are in the [non]terminal_symbols set
  set<Symbol>::iterator iter_sy_f;
  for (vector<Symbol>::const_iterator iter_sy = left.begin();
       iter_sy != left.end(); ++iter_sy)
    if (isSymbolNonterminal(*iter_sy)) {
      iter_sy_f = nonterminal_symbols.find(*iter_sy);
      if (iter_sy_f == nonterminal_symbols.end()) {
	setError(E_G_NONTERMINAL_SYMBOL_BAD); return false;
      }
    } else {
      iter_sy_f = terminal_symbols.find(*iter_sy);
      if (iter_sy_f == terminal_symbols.end()) {
	setError(E_G_TERMINAL_SYMBOL_BAD); return false;
      }
    }
  for (vector<Symbol>::const_iterator iter_sy = right.begin();
       iter_sy != right.end(); ++iter_sy)
    if (isSymbolNonterminal(*iter_sy)) {
      iter_sy_f = nonterminal_symbols.find(*iter_sy);
      if (iter_sy_f == nonterminal_symbols.end()) {
	setError(E_G_NONTERMINAL_SYMBOL_BAD); return false;
      }
    } else {
      iter_sy_f = terminal_symbols.find(*iter_sy);
      if (iter_sy_f == terminal_symbols.end()) {
	setError(E_G_TERMINAL_SYMBOL_BAD); return false;
      }
    }
  Rule rule;
  rule.left = left;
  rule.right = right;
  return removeRule(rule);
}

inline set<Rule>
Grammar::getRules()
{
  return rules;
}

bool
Grammar::setStartSymbol(NonterminalSymbol symbol)
{
  // check if the symbol is in the nonterminal_symbols set
  set<NonterminalSymbol>::iterator iter_sy = nonterminal_symbols.find(symbol);
  if (iter_sy == nonterminal_symbols.end()) {
    setError(E_G_NONTERMINAL_SYMBOL_BAD); return false;
  }
  start_symbol = symbol;
  return true;
}

inline NonterminalSymbol
Grammar::getStartSymbol()
{
  return start_symbol;
}

Grammar::t_classification
Grammar::getClassification()
{
  if (isRegular()) return C_REGULAR;
  if (isContextfree()) return C_CONTEXTFREE;
  if (isContext()) return C_CONTEXT;
  return C_BOUNDLESS;
}

inline bool
Grammar::isSymbolNonterminal(Symbol symbol)
{
  return 
    ((symbol >= NonterminalSymbol_MIN) && (symbol <= NonterminalSymbol_MAX));
}

inline bool
Grammar::isSymbolTerminal(Symbol symbol)
{
  return ((symbol >= TerminalSymbol_MIN) && (symbol <= TerminalSymbol_MAX));
}

bool
Grammar::isRegular()
{
  vector<Symbol> str;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) {
    str = iter_r->left;
    if (str.size() != 1) return false;
    if (!isSymbolNonterminal(str.front())) return false;
    str = iter_r->right;
    if ((str.size() != 1) && (str.size() != 2)) {
      if (!str.empty()) return false;
      else if (iter_r->left.front() != start_symbol) return false;
    }
    if (!isSymbolTerminal(str.front())) 
      if (iter_r->left.front() != start_symbol) return false;
    if (str.size() == 2) if (!isSymbolNonterminal(str[1])) return false;
  }
  return true;
}

bool
Grammar::isContextfree()
{
  vector<Symbol> str;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) {
    str = iter_r->left;
    if (str.size() != 1) return false;
    if (!isSymbolNonterminal(str.front())) return false;
  }
  return true;
}

bool
Grammar::isContext()
{
  vector<Symbol> str;
  vector<Symbol>::const_iterator iter_sy;
  for (set<Rule>::const_iterator iter_r = rules.begin();
       iter_r != rules.end(); ++iter_r) {
    str = iter_r->left;
    // left part must include at least one nonterminal symbol
    for (iter_sy = str.begin(); iter_sy != str.end(); ++iter_sy) 
      if (isSymbolNonterminal(*iter_sy)) break;
    if (iter_sy == str.end()) return false;
    if (iter_r->right.empty()) 
      if (iter_r->left.front() != start_symbol) return false;
    if (str.size() > iter_r->right.size()) return false;
  }
  return true;
}

NonterminalSymbol
Grammar::toExtended()
{
  if (start_symbol == NonterminalSymbol_NULL) return start_symbol;
  NonterminalSymbol old_start_symbol = start_symbol;
  NonterminalSymbol new_start_symbol = addNonterminalSymbol();
  // we don't (may not) want checks
  Rule rule;
  rule.left = vector<Symbol>(1, new_start_symbol);
  rule.right = vector<Symbol>(1, old_start_symbol);
  rules.insert(rule);
  rule.right = vector<Symbol>();
  rules.insert(rule);
  /*
  addRule
    (vector<Symbol>(1, new_start_symbol), vector<Symbol>(1, old_start_symbol));
  addRule(vector<Symbol>(1, new_start_symbol), vector<Symbol>());
  */
  setStartSymbol(new_start_symbol);
  return old_start_symbol;
}
