/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "error.h"
#include "grammar_l.h"

const char string_space = ' ';
const char string_lbracket = '(';
const char string_rbracket = ')';

Grammar_l::Grammar_l()
{
  grammar = new Grammar;
}

Grammar_l::~Grammar_l()
{
  delete grammar;
  grammar = NULL;
}

string
Grammar_l::addNonterminalSymbol(string name)
{
  string result;
  // check if we already don't have nonteminal symbol with such name
  map<NonterminalSymbol, NonterminalSymbol_l>::iterator
    iter_map = nonterminal_symbols.begin();
  for (; iter_map != nonterminal_symbols.end(); ++iter_map)
    if (iter_map->second.getName() == name) break;
  if (iter_map != nonterminal_symbols.end()) {
    setError(E_G_NONTERMINAL_SYMBOL_NAME_EXISTS); return result;
  }
  NonterminalSymbol new_symbol = grammar->addNonterminalSymbol();
  if (new_symbol == NonterminalSymbol_NULL) return result;
  NonterminalSymbol_l new_symbol_l(name);
  nonterminal_symbols[new_symbol] = new_symbol_l;
  // return look
  result = 
    new_symbol_l.getLook(new_symbol, getNonterminalSymbolFlags(new_symbol));
  return result;
}

NonterminalSymbol_l::t_flags
Grammar_l::getNonterminalSymbolFlags(NonterminalSymbol symbol)
{
  NonterminalSymbol_l::t_flags result = (NonterminalSymbol_l::t_flags)0;
  if (grammar->getStartSymbol() == symbol) 
    result = 
      (NonterminalSymbol_l::t_flags)(result | NonterminalSymbol_l::F_START);
  return result;
}

bool
Grammar_l::removeNonterminalSymbol(string symbol_id)
{
  NonterminalSymbol symbol = string_to_Symbol(symbol_id);
  if (grammar->removeNonterminalSymbol(symbol)){ 
    nonterminal_symbols.erase(symbol); return true;
  }
  return false;
}

Symbol
Grammar_l::string_to_Symbol(string symbol_id)
{
  Symbol result = 0;
  int mult_factor = 1;
  reverse(symbol_id.begin(), symbol_id.end());
  for (string::const_iterator iter_s = symbol_id.begin();
       iter_s != symbol_id.end(); ++iter_s, mult_factor *= 10)
    result += (*iter_s - '0') * mult_factor;
  return result;
}

string
Grammar_l::editNonterminalSymbol(string symbol_id, string new_name)
{
  string result;
  NonterminalSymbol symbol = string_to_Symbol(symbol_id);
  map<NonterminalSymbol, NonterminalSymbol_l>::iterator 
    iter_sy_f = nonterminal_symbols.find(symbol);
  if (iter_sy_f == nonterminal_symbols.end()) { 
    setError(E_G_NONTERMINAL_SYMBOL_BAD); return result;
  }
  // check if we already don't have nonterminal symbol with such name
  iter_sy_f = nonterminal_symbols.begin();
  for (; iter_sy_f != nonterminal_symbols.end(); ++iter_sy_f)
    if (iter_sy_f->second.getName() == new_name) break;
  if (iter_sy_f != nonterminal_symbols.end()) { 
    setError(E_G_NONTERMINAL_SYMBOL_NAME_EXISTS); return result;
  }
  NonterminalSymbol_l new_symbol_l(new_name);
  nonterminal_symbols[symbol] = new_symbol_l;
  // return look
  result = new_symbol_l.getLook(symbol, getNonterminalSymbolFlags(symbol));
  return result;
}

string
Grammar_l::addTerminalSymbol(string name)
{
  string result;
  // check if we already don't have terminal symbol with such name
  map<TerminalSymbol, TerminalSymbol_l>::iterator
    iter_sy = terminal_symbols.begin();
  for (; iter_sy != terminal_symbols.end(); ++iter_sy) 
    if (iter_sy->second.getName() == name) break;
  if (iter_sy != terminal_symbols.end()) {
    setError(E_G_TERMINAL_SYMBOL_NAME_EXISTS); return result;
  }
  TerminalSymbol new_symbol = grammar->addTerminalSymbol();
  if (new_symbol == TerminalSymbol_NULL) return result;
  TerminalSymbol_l new_symbol_l(name);
  terminal_symbols[new_symbol] = new_symbol_l;
  // return look
  result = new_symbol_l.getLook(new_symbol);
  return result;
}

bool
Grammar_l::removeTerminalSymbol(string symbol_id)
{
  TerminalSymbol symbol = string_to_Symbol(symbol_id);
  if (grammar->removeTerminalSymbol(symbol)) {
    terminal_symbols.erase(symbol); return true;
  }
  return false;
}

string
Grammar_l::editTerminalSymbol(string symbol_id, string new_name)
{
  string result;
  TerminalSymbol symbol = string_to_Symbol(symbol_id);
  map<TerminalSymbol, TerminalSymbol_l>::iterator 
    iter_sy_f = terminal_symbols.find(symbol);
  if (iter_sy_f == terminal_symbols.end()) { 
    setError(E_G_TERMINAL_SYMBOL_BAD); return result;
  }
  // check if we already don't have terminal symbol with such name
  iter_sy_f = terminal_symbols.begin();
  for (; iter_sy_f != terminal_symbols.end(); ++iter_sy_f)
    if (iter_sy_f->second.getName() == new_name) break;
  if (iter_sy_f != terminal_symbols.end()) { 
    setError(E_G_TERMINAL_SYMBOL_NAME_EXISTS); return result;
  }
  TerminalSymbol_l new_symbol_l(new_name);
  terminal_symbols[symbol] = new_symbol_l;
  // return look
  result = new_symbol_l.getLook(symbol);
  return result;
}

bool
Grammar_l::setStartSymbol(string symbol_id)
{
  return grammar->setStartSymbol(string_to_Symbol(symbol_id));
}

bool
Grammar_l::addRule(string left, string right)
{
  vector<Symbol> lft, rght;
  string symbol;
  // parse symbol strings
  for (string::const_iterator iter_s = left.begin(); iter_s != left.end();
       ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      lft.push_back(string_to_Symbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) lft.push_back(string_to_Symbol(symbol));
  symbol.clear();
  for (string::const_iterator iter_s = right.begin(); iter_s != right.end();
       ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      rght.push_back(string_to_Symbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) rght.push_back(string_to_Symbol(symbol));
  return grammar->addRule(lft, rght);
}

bool
Grammar_l::removeRule(string left, string right)
{
  vector<Symbol> lft, rght;
  string symbol;
  // parse symbol strings
  for (string::const_iterator iter_s = left.begin(); iter_s != left.end();
       ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      lft.push_back(string_to_Symbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) lft.push_back(string_to_Symbol(symbol));
  symbol.clear();
  for (string::const_iterator iter_s = right.begin(); iter_s != right.end();
       ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      rght.push_back(string_to_Symbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) rght.push_back(string_to_Symbol(symbol));
  return grammar->removeRule(lft, rght);
}

string
Grammar_l::getLook()
{
  string result;
  result += string_lbracket; result += "grammar"; result += string_space;
  // NonterminalSymbol looks
  result += string_lbracket; result += "nonterminal_symbols";
  set<NonterminalSymbol> nt_symbols = grammar->getNonterminalSymbols();
  for (set<NonterminalSymbol>::const_iterator iter_sy = nt_symbols.begin();
       iter_sy != nt_symbols.end(); ++iter_sy)
    result += string_space +
      nonterminal_symbols[*iter_sy].getLook
      (*iter_sy, getNonterminalSymbolFlags(*iter_sy));
  result += string_rbracket; result += string_space;
  // TerminalSymbol looks
  result += string_lbracket; result += "terminal_symbols";
  set<TerminalSymbol> t_symbols = grammar->getTerminalSymbols();
  for (set<TerminalSymbol>::const_iterator iter_sy = t_symbols.begin();
       iter_sy != t_symbols.end(); ++iter_sy)
    result += string_space + terminal_symbols[*iter_sy].getLook(*iter_sy);
  result += string_rbracket; result += string_space;
  // rule looks
  result += string_lbracket; result += "rules";
  set<Rule> rules = grammar->getRules();
  for (set<Rule>::const_iterator iter_r = rules.begin(); iter_r != rules.end();
       ++iter_r) {
    result += string_space; result += string_lbracket; result += "rule";
    result += string_space; result += string_lbracket; result += "left";
    for (vector<Symbol>::const_iterator iter_sy = iter_r->left.begin(); 
         iter_sy != iter_r->left.end(); ++iter_sy)
      if (grammar->isSymbolNonterminal(*iter_sy))
	result += string_space + 
	  nonterminal_symbols[*iter_sy].getLook
	  (*iter_sy, getNonterminalSymbolFlags(*iter_sy));
      else
	result += string_space + terminal_symbols[*iter_sy].getLook(*iter_sy);
    result += string_rbracket;
    result += string_space; result += string_lbracket; result += "right";
    for (vector<Symbol>::const_iterator iter_sy = iter_r->right.begin(); 
         iter_sy != iter_r->right.end(); ++iter_sy)
      if (grammar->isSymbolNonterminal(*iter_sy))
	result += string_space + 
	  nonterminal_symbols[*iter_sy].getLook
	  (*iter_sy, getNonterminalSymbolFlags(*iter_sy));
      else
	result += string_space + terminal_symbols[*iter_sy].getLook(*iter_sy);
    result += string_rbracket;
    result += string_rbracket;
  }
  result += string_rbracket; result += string_rbracket;
  return result;
}

string
Grammar_l::getClassification()
{
  if (grammar->getClassification() == Grammar::C_REGULAR) return "REGULAR";
  if (grammar->getClassification() == Grammar::C_CONTEXTFREE)
    return "CONTEXTFREE";
  if (grammar->getClassification() == Grammar::C_CONTEXT) return "CONTEXT";
  return "BOUNDLESS";
}

void
Grammar_l::toExtended()
{
  NonterminalSymbol old_start_symbol = grammar->toExtended();
  string new_name = nonterminal_symbols[old_start_symbol].getName() + "'";
  nonterminal_symbols[grammar->getStartSymbol()] = 
    NonterminalSymbol_l(new_name);
}

const Grammar_l &
Grammar_l::operator=(const Grammar_l &g)
{
  if (this != &g) {
    nonterminal_symbols = g.nonterminal_symbols;
    terminal_symbols = g.terminal_symbols;
    *grammar = *(g.grammar);
  }
  return *this;
}
