/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "error.h"
#include "stack_automata.h"

StackAutomata::StackAutomata()
{
  start_stack_symbol = StackSymbol_NULL;
}

StackAutomata::~StackAutomata()
{

}

bool
StackAutomata::removeState(State state)
{
  set<State>::iterator iter_st_f = states.find(state);
  if (iter_st_f == states.end()) { setError(E_A_STATE_BAD); return false; }
  // erase also transitions which contain the state
  set<StackTransition> trans = find_transitions_with_state(state);
  for (set<StackTransition>::const_iterator iter_t = trans.begin();
       iter_t != trans.end(); ++iter_t) removeTransition(*iter_t);
  // if we're erasing start state we must set new one
  if (state == start_state) {
    if (states.size() != 1) {
      if (state == *states.begin()) start_state = *++states.begin();
      else start_state = *states.begin();
    }
    else start_state = State_NULL;
  }
  // if we're erasing configuration state we must set new one (start)
  if (state == c_state) c_state = start_state;
  states.erase(state);
  return true;
}

set<StackTransition>
StackAutomata::find_transitions_with_state(State state)
{
  set<StackTransition> result;
  for (set<StackTransition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t) {
    if (iter_t->from_state == state) { result.insert(*iter_t); continue; }
    if (iter_t->to_state == state) result.insert(*iter_t);
  }
  return result;
}

bool
StackAutomata::removeEntrySymbol(EntrySymbol symbol)
{
  set<EntrySymbol>::iterator iter_sy_f = entry_symbols.find(symbol);
  if (iter_sy_f == entry_symbols.end()) { 
    setError(E_A_ENTRY_SYMBOL_BAD); return false;
  }
  // erase also transitions which contain the symbol
  set<StackTransition> trans = find_transitions_with_entry_symbol(symbol);
  for (set<StackTransition>::const_iterator iter_t = trans.begin();
       iter_t != trans.end(); ++iter_t) removeTransition(*iter_t);
  entry_symbols.erase(symbol);
  return true;
}

set<StackTransition>
StackAutomata::find_transitions_with_entry_symbol(EntrySymbol symbol)
{
  set<StackTransition> result;
  for (set<StackTransition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t)
    if (iter_t->entry_symbol == symbol) result.insert(*iter_t);
  return result;
}

StackSymbol
StackAutomata::addStackSymbol()
{
  StackSymbol new_symbol = StackSymbol_MIN;
  for (set<StackSymbol>::const_iterator iter_sy = stack_symbols.begin();
       iter_sy != stack_symbols.end(); ++iter_sy, ++new_symbol)
    if (*iter_sy != new_symbol) break;
  if (new_symbol > StackSymbol_MAX) { 
    setError(E_AS_STACK_SYMBOL_MAX); return StackSymbol_NULL;
  }
  stack_symbols.insert(new_symbol);
  // if there's only one stack symbol it's start and configuration stack
  if (stack_symbols.size() == 1) {
    start_stack_symbol = new_symbol;
    c_stack.push_back(start_stack_symbol);
  }
  return new_symbol;
}

bool
StackAutomata::removeStackSymbol(StackSymbol symbol)
{
  set<StackSymbol>::iterator iter_sy_f = stack_symbols.find(symbol);
  if (iter_sy_f == stack_symbols.end()) { 
    setError(E_AS_STACK_SYMBOL_BAD); return false;
  }
  // erase also transitions which contain the symbol
  set<StackTransition> trans = find_transitions_with_stack_symbol(symbol);
  for (set<StackTransition>::const_iterator iter_t = trans.begin();
       iter_t != trans.end(); ++iter_t) removeTransition(*iter_t);
  // if we're erasing start stack symbol we must set new one
  if (symbol == start_stack_symbol) {
    if (stack_symbols.size() != 1) {
      if (symbol == *stack_symbols.begin()) 
	start_stack_symbol = *++stack_symbols.begin();
      else start_stack_symbol = *stack_symbols.begin();
    }
    else start_stack_symbol = StackSymbol_NULL;
  }
  // if we're erasing configuration stack symbol we must remove it
  // (ALL occurrences)
  bool erased;
  do {
    erased = false;
    for (vector<StackSymbol>::iterator iter_sy = c_stack.begin();
	 iter_sy != c_stack.end(); ++iter_sy) 
      if (*iter_sy == symbol) { c_stack.erase(iter_sy); erased = true; break; }
  } while (erased);
  stack_symbols.erase(symbol);
  return true;
}

set<StackTransition>
StackAutomata::find_transitions_with_stack_symbol(StackSymbol symbol)
{
  set<StackTransition> result;
  vector<StackSymbol> stk;
  for (set<StackTransition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t) {
    stk = iter_t->from_stack;
    for (vector<StackSymbol>::const_iterator iter_sy = stk.begin();
	 iter_sy != stk.end(); ++iter_sy) 
      if (*iter_sy == symbol) { result.insert(*iter_t); break; }
    stk = iter_t->to_stack;
    for (vector<StackSymbol>::const_iterator iter_sy = stk.begin();
	 iter_sy != stk.end(); ++iter_sy) 
      if (*iter_sy == symbol) { result.insert(*iter_t); break; }
  }    
  return result;
}

inline set<StackSymbol>
StackAutomata::getStackSymbols()
{
  return stack_symbols;
}

bool
StackAutomata::removeTransition(StackTransition transition)
{
  set<StackTransition>::iterator iter_t_f = transitions.find(transition);
  if (iter_t_f == transitions.end()) { 
    setError(E_A_TRANSITION_BAD); return false;
  }
  transitions.erase(transition);
  return true;
}

inline set<StackTransition>
StackAutomata::getTransitions()
{
  return transitions;
}

bool
StackAutomata::setStartStackSymbol(StackSymbol symbol)
{
  // check if the symbol is in the stack symbol set
  set<StackSymbol>::iterator iter_sy = stack_symbols.find(symbol);
  if (iter_sy == stack_symbols.end()) { 
    setError(E_AS_STACK_SYMBOL_BAD); return false;
  }
  start_stack_symbol = symbol;
  return true;
}

inline StackSymbol
StackAutomata::getStartStackSymbol()
{
  return start_stack_symbol;
}

bool
StackAutomata::addTransition(StackTransition &transition, State from_state,
			     EntrySymbol symbol, State to_state,
			     vector<StackSymbol> from_stack,
			     vector<StackSymbol> to_stack)
{
  // check if both states are in the states set
  set<State>::iterator iter_st = states.find(from_state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  iter_st = states.find(to_state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  // check if the symbol is in the entry_symbols set
  // CHOICE: if not, add it - now NO
  if (symbol != EntrySymbol_NULL) {
    set<EntrySymbol>::iterator iter_sy = entry_symbols.find(symbol);
    if (iter_sy == entry_symbols.end()) { 
      setError(E_A_ENTRY_SYMBOL_BAD); return false;
    }
  }
  // check if all stack symbols in both stacks are in the stack_symbols set
  // CHOICE: if not, add it - now NO
  set<StackSymbol>::iterator iter_sy_f;
  for (vector<StackSymbol>::const_iterator iter_sy = from_stack.begin();
       iter_sy != from_stack.end(); ++iter_sy) {
    iter_sy_f = stack_symbols.find(*iter_sy);
    if (iter_sy_f == stack_symbols.end()) { 
      setError(E_AS_STACK_SYMBOL_BAD); return false;
    }
  }
  for (vector<StackSymbol>::const_iterator iter_sy = to_stack.begin();
       iter_sy != to_stack.end(); ++iter_sy) {
    iter_sy_f = stack_symbols.find(*iter_sy);
    if (iter_sy_f == stack_symbols.end()) {
      setError(E_AS_STACK_SYMBOL_BAD); return false;
    }
  }
  transition.from_state = from_state;
  transition.to_state = to_state;
  transition.entry_symbol = symbol;
  transition.from_stack = from_stack;
  transition.to_stack = to_stack;
  transitions.insert(transition);
  return true;
}

bool
StackAutomata::addTransition(State from_state, EntrySymbol symbol,
			     State to_state, vector<StackSymbol> from_stack,
			     vector<StackSymbol> to_stack)
{
  StackTransition tran;
  return 
    addTransition(tran, from_state, symbol, to_state, from_stack, to_stack);
}

bool
StackAutomata::removeTransition(State from_state, EntrySymbol symbol,
				State to_state, 
				vector<StackSymbol> from_stack,
				vector<StackSymbol> to_stack)
{
  // check if both states are in the states set
  set<State>::iterator iter_st = states.find(from_state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  iter_st = states.find(to_state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  // check if the symbol is in the entry_symbols set
  if (symbol != EntrySymbol_NULL) {
    set<EntrySymbol>::iterator iter_sy = entry_symbols.find(symbol);
    if (iter_sy == entry_symbols.end()) {
      setError(E_A_ENTRY_SYMBOL_BAD); return false;
    }
  }
  // check if all stack symbols in both stacks are in the stack_symbols set
  set<StackSymbol>::iterator iter_sy_f;
  for (vector<StackSymbol>::const_iterator iter_sy = from_stack.begin();
       iter_sy != from_stack.end(); ++iter_sy) {
    iter_sy_f = stack_symbols.find(*iter_sy);
    if (iter_sy_f == stack_symbols.end()) {
      setError(E_AS_STACK_SYMBOL_BAD); return false;
    }
  }
  for (vector<StackSymbol>::const_iterator iter_sy = to_stack.begin();
       iter_sy != to_stack.end(); ++iter_sy) {
    iter_sy_f = stack_symbols.find(*iter_sy);
    if (iter_sy_f == stack_symbols.end()) {
      setError(E_AS_STACK_SYMBOL_BAD); return false;
    }
  }
  StackTransition transition;
  transition.from_state = from_state;
  transition.to_state = to_state;
  transition.entry_symbol = symbol;
  transition.from_stack = from_stack;
  transition.to_stack = to_stack;
  return removeTransition(transition);
}

bool
StackAutomata::setConfiguration(State state, vector<EntrySymbol> sentence,
				vector<StackSymbol> stack)
{
  // check if the state is in the states set
  set<State>::iterator iter_st = states.find(state);
  if (iter_st == states.end()) { setError(E_A_STATE_BAD); return false; }
  // check if all symbols in the sentence are in the entry_symbols set
  // CHOICE: if not, add it - now NO
  set<EntrySymbol>::iterator iter_sy_f;
  for (vector<EntrySymbol>::const_iterator iter_sy = sentence.begin();
       iter_sy != sentence.end(); ++iter_sy) {
    iter_sy_f = entry_symbols.find(*iter_sy);
    if (iter_sy_f == entry_symbols.end()) {
      setError(E_A_ENTRY_SYMBOL_BAD); return false;
    }
  }
  // check if all symbols in the stack are in the stack_symbols set
  // CHOICE: if not, add it - now NO
  set<StackSymbol>::iterator iter_sy_f_1;
  for (vector<StackSymbol>::const_iterator iter_sy = stack.begin();
       iter_sy != stack.end(); ++iter_sy) {
    iter_sy_f_1 = stack_symbols.find(*iter_sy);
    if (iter_sy_f_1 == stack_symbols.end()) {
      setError(E_AS_STACK_SYMBOL_BAD); return false;
    }
  }
  c_state = state;
  c_sentence = sentence;
  c_stack = stack;
  return true;
}

inline void
StackAutomata::getConfiguration(State& state, vector<EntrySymbol>& sentence,
				vector<StackSymbol>& stack)
{
  state = c_state;
  sentence = c_sentence;
  stack = c_stack;
}

inline set<vector<StackSymbol> >
StackAutomata::getFromStacks()
{
  return _getFromStacks(c_state, c_stack);
}

set<vector<StackSymbol> >
StackAutomata::_getFromStacks(State state, vector<StackSymbol> stack)
{
  set<vector<StackSymbol> > result;
  for (set<StackTransition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t)
    if (iter_t->from_state == state) { 
      // check if from_stack can be popped from c_stack
      vector<StackSymbol> stk = stack, from_stk = iter_t->from_stack;
      while ((!from_stk.empty()) && (!stk.empty())) {
	if (from_stk.back() == stk.back()) { 
	  from_stk.pop_back(); stk.pop_back();
	}
	// if stack symbol can't be popped, transition can't be executed
	else break;
      }
      if (from_stk.empty()) result.insert(iter_t->from_stack);
    }
  return result;
}

map<State, vector<StackSymbol> >
StackAutomata::getTransitionDestinations(vector<StackSymbol> from_stack)
{
  map<State, vector<StackSymbol> > result, e_trans;
  e_trans = _getTransitionDestinations(c_state, EntrySymbol_NULL, from_stack);
  if (!c_sentence.empty())
    result =
      _getTransitionDestinations(c_state, c_sentence.front(), from_stack);
  result.insert(e_trans.begin(), e_trans.end());
  return result;
}

map<State, vector<StackSymbol> >
StackAutomata::_getTransitionDestinations(State state, EntrySymbol symbol,
					  vector<StackSymbol> stack)
{
  map<State, vector<StackSymbol> > result;
  if (state == State_NULL) return result;
  // we want to_state and to_stack of transitions which have
  // from_state == state, entry_symbol == symbol and from_stack == stack
  for (set<StackTransition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t)
    if ((iter_t->from_state == state) &&
        (iter_t->entry_symbol == symbol) &&
	(iter_t->from_stack == stack))
      result[iter_t->to_state] = iter_t->to_stack;
  return result;
}

inline bool
StackAutomata::doTransition(State state, vector<StackSymbol> from_stack,
			    vector<StackSymbol> to_stack)
{
  return 
    _doTransition(c_state, state, c_sentence, from_stack, to_stack, c_stack);
}

bool
StackAutomata::_doTransition(State& from_state, State to_state,
			     vector<EntrySymbol>& sentence,
			     vector<StackSymbol> from_stack,
			     vector<StackSymbol> to_stack,
			     vector<StackSymbol> &stack)
{
  map<State, vector<StackSymbol> > dests, e_trans;
  e_trans =
    _getTransitionDestinations(from_state, EntrySymbol_NULL, from_stack);
  if (!sentence.empty())
    dests = 
      _getTransitionDestinations(from_state, sentence.front(), from_stack);
  bool e_tran = false;
  // check if the pair to_state and to_stack is among dests or e_trans
  map<State, vector<StackSymbol> >::const_iterator iter_map = dests.begin();
  for (; iter_map != dests.end(); ++iter_map)
    if ((iter_map->first == to_state) && (iter_map->second == to_stack)) break;
  if (iter_map == dests.end()) {
    for (iter_map = e_trans.begin(); iter_map != e_trans.end(); ++iter_map)
      if ((iter_map->first == to_state) && (iter_map->second == to_stack))
	break;
    if (iter_map == e_trans.end()) { 
      setError(E_AS_STATE_STACK_CANT_REACH); return false;
    }
    else e_tran = true;
  }
  // pop from_stack from stack and push to_stack
  vector<StackSymbol> stk = stack, from_stk = from_stack;
  while (!from_stk.empty()) {
    if (from_stk.back() == stk.back()) { from_stk.pop_back(); stk.pop_back(); }
    // if stack symbol can't be popped, transition can't be executed
    else { setError(E_AS_STACK_CANT_POP); return false; }
  }
  for (vector<StackSymbol>::const_iterator iter_sy = to_stack.begin();
       iter_sy != to_stack.end(); ++iter_sy) stk.push_back(*iter_sy);
  stack = stk;
  // move to the to_state
  from_state = to_state;
  // erase the first symbol of the sentence
  if (!e_tran) sentence.erase(sentence.begin());
  return true;
}

bool
StackAutomata::checkSentence(vector<EntrySymbol> sentence)
{
  vector<StackSymbol> start_stack, stack;
  start_stack.push_back(start_stack_symbol);
  set<vector<StackSymbol> > 
    from_stacks = _getFromStacks(start_state, start_stack);
  // check for all from_stacks
  for (set<vector<StackSymbol> >::const_iterator iter_v = from_stacks.begin();
       iter_v != from_stacks.end(); ++iter_v) {
    stack = start_stack;
    if (_checkSentence(start_state, sentence, *iter_v, stack)) return true;
  }
  return false;
}

bool
StackAutomata::_checkSentence(State from_state, vector<EntrySymbol> sentence,
			      vector<StackSymbol> from_stack,
			      vector<StackSymbol>& stack)
{
  map<State, vector<StackSymbol> > dests, e_trans;
  e_trans =
    _getTransitionDestinations(from_state, EntrySymbol_NULL, from_stack);
  if (!sentence.empty())
    dests = 
      _getTransitionDestinations(from_state, sentence.front(), from_stack);
  dests.insert(e_trans.begin(), e_trans.end());
  State new_state;
  vector<EntrySymbol> rest_of_sentence = sentence;
  set<vector<StackSymbol> > new_stacks;
  for (map<State, vector<StackSymbol> >::const_iterator
	 iter_map = dests.begin(); iter_map != dests.end(); ++iter_map) {
    // we must have the same conditions for each *iter_map
    new_state = from_state;
    rest_of_sentence = sentence;
    _doTransition
      (new_state, iter_map->first, rest_of_sentence, from_stack,
       iter_map->second, stack);
    new_stacks = _getFromStacks(new_state, stack);
    for (set<vector<StackSymbol> >::const_iterator
	   iter_v = new_stacks.begin(); iter_v != new_stacks.end(); ++iter_v)
      // check the rest of the sentence for all new from stacks
      if (_checkSentence(new_state, rest_of_sentence, *iter_v, stack))
	return true;
    if (rest_of_sentence.empty())
      if (isEmptyStackChecking()) { if (stack.empty()) return true; }
      else {
	set<State>::iterator iter_st = end_states.find(new_state);
	if (iter_st != end_states.end()) return true;
      }
  }
  // check if all entry symbols were accepted
  if (rest_of_sentence.empty()) {
    if (isEmptyStackChecking()) return stack.empty();
    set<State>::iterator iter_st = end_states.find(new_state);
    if (iter_st != end_states.end()) return true;
  }
  return false;
}

bool
StackAutomata::isDeterministic()
{
  for (set<StackTransition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t) {
    if (_getTransitionDestinations
	(iter_t->from_state, iter_t->entry_symbol, iter_t->from_stack).size()
	> 1) return false;
    for (set<StackTransition>::const_iterator iter_t_1 = transitions.begin();
       iter_t_1 != transitions.end(); ++iter_t_1)
      if ((iter_t != iter_t_1) &&
	  (iter_t->from_state == iter_t_1->from_state)) {
	if ((iter_t_1->entry_symbol == EntrySymbol_NULL) ||
	    (iter_t_1->entry_symbol == iter_t->entry_symbol))
	  if (isPrefix12or21(iter_t->from_stack, iter_t_1->from_stack))
	    return false;
      }
  }
  return true;
}

bool
StackAutomata::isPrefix12or21(vector<StackSymbol> stack1,
			      vector<StackSymbol> stack2)
{
  while ((!stack1.empty()) && (!stack2.empty())) {
    if (stack1.back() == stack2.back()) { 
      stack1.pop_back(); stack2.pop_back();
    }
    // if last stack symbols differ, stop comparing
    else break;
  }
  if ((stack1.empty()) || (stack2.empty())) return true; else return false;
}

bool
StackAutomata::isTopStackWorking()
{
  for (set<StackTransition>::const_iterator iter_t = transitions.begin();
       iter_t != transitions.end(); ++iter_t)
    if (iter_t->from_stack.size() != 1) return false;
  return true;
}

void
StackAutomata::toTopStackWorking(State &old_start_state,
				 StackSymbol &old_start_stack_symbol,
				 map<State, vector<State> > &new_states)
{
  if (isTopStackWorking()) return;
  newStart(old_start_state, old_start_stack_symbol);
  new_states.clear();
  vector<StackSymbol> from_stack, to_stack;
  State new_state1, new_state2;
  bool remove_tran;
  // change transitions
  set<StackTransition> trans = transitions;
  for (set<StackTransition>::const_iterator iter_t = trans.begin();
       iter_t != trans.end(); ++iter_t) {
    remove_tran = false;
    from_stack = iter_t->from_stack;
    if (from_stack.empty()) {
      for (set<StackSymbol>::const_iterator iter_sy = stack_symbols.begin();
	   iter_sy != stack_symbols.end(); ++iter_sy) {
	to_stack = iter_t->to_stack;
	to_stack.insert(to_stack.begin(), *iter_sy);
	addTransition(iter_t->from_state, iter_t->entry_symbol,
		      iter_t->to_state, vector<StackSymbol>(1, *iter_sy),
		      to_stack);
      }
      remove_tran = true;
    }
    if (from_stack.size() > 1) {
      for (vector<StackSymbol>::const_iterator iter_sy = from_stack.begin();
	   iter_sy != from_stack.end(); ++iter_sy)
	if (iter_sy == from_stack.begin()) {
	  new_state1 = addState();
	  new_states[iter_t->from_state].push_back(new_state1);
	  addTransition(iter_t->from_state, iter_t->entry_symbol, new_state1,
			vector<StackSymbol>(1, *iter_sy),
			vector<StackSymbol>());
	}
	else if (*iter_sy != from_stack.back()) {
	  new_state2 = addState();
	  new_states[iter_t->from_state].push_back(new_state2);
	  addTransition(new_state1, EntrySymbol_NULL, new_state2,
			vector<StackSymbol>(1, *iter_sy),
			vector<StackSymbol>());
	  new_state1 = new_state2;
	}
	else
	  addTransition(new_state1, EntrySymbol_NULL, iter_t->to_state,
			vector<StackSymbol>(1, *iter_sy), iter_t->to_stack);
      remove_tran = true;
    }
    // remove transition
    if (remove_tran) transitions.erase(*iter_t);
  }
}
      
inline bool
StackAutomata::isEmptyStackChecking()
{
  return end_states.empty();
}

void
StackAutomata::toEmptyStackChecking(State &old_start_state,
				    StackSymbol &old_start_stack_symbol,
				    State &new_state)
{
  if (isEmptyStackChecking()) return;
  if ((start_state == State_NULL) || (start_stack_symbol == StackSymbol_NULL))
    return;
  newStart(old_start_state, old_start_stack_symbol);
  // state in which we empty the stack
  new_state = addState();
  for (set<StackSymbol>::const_iterator iter_sy = stack_symbols.begin();
       iter_sy != stack_symbols.end(); ++iter_sy)
    for (set<State>::const_iterator iter_st = states.begin(); 
	 iter_st != states.end(); ++iter_st) {
      // empty the stack in end state or new_state
      set<State>::iterator iter_st_f = end_states.find(*iter_st);
      if ((iter_st_f != end_states.end()) || (*iter_st == new_state))
	addTransition(*iter_st, EntrySymbol_NULL, new_state,
		      vector<StackSymbol>(1, *iter_sy), vector<StackSymbol>());
    }
  end_states.clear();
}

void
StackAutomata::newStart(State &old_start_state,
			StackSymbol &old_start_stack_symbol)
{
  old_start_state = start_state;
  old_start_stack_symbol = start_stack_symbol;
  State new_start_state = addState();
  StackSymbol new_start_stack_symbol = addStackSymbol();
  vector<StackSymbol> from_stack = 
    vector<StackSymbol>(1, new_start_stack_symbol);
  vector<StackSymbol> to_stack = from_stack;
  to_stack.push_back(start_stack_symbol);
  addTransition(new_start_state, EntrySymbol_NULL, start_state, from_stack,
		to_stack);
  setStartState(new_start_state);
  setStartStackSymbol(new_start_stack_symbol);
}

void
StackAutomata::toEndStatesChecking(State &old_start_state,
				   StackSymbol &old_start_stack_symbol,
				   State &end_state)
{
  if (!isEmptyStackChecking()) return;
  if ((start_state == State_NULL) || (start_stack_symbol == StackSymbol_NULL))
    return;
  newStart(old_start_state, old_start_stack_symbol);
  // end state
  end_state = addState();
  for (set<State>::const_iterator iter_st = states.begin(); 
       iter_st != states.end(); ++iter_st)
    // if the (original) stack is empty move to end_state
    if (*iter_st != end_state)
      addTransition(*iter_st, EntrySymbol_NULL, end_state,
		    vector<StackSymbol>(1, start_stack_symbol),
		    vector<StackSymbol>());
  end_states.insert(end_state);
}
