/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// class of the stack automata

#ifndef __STACK_AUTOMATA_H__
#define __STACK_AUTOMATA_H__

#include "automata.h"
#include "stack_automata_elements.h"
#include <map>

class StackAutomata : public Automata {
 protected:
  set<StackSymbol> stack_symbols;
  set<StackTransition> transitions;
  StackSymbol start_stack_symbol;
  vector<StackSymbol> c_stack; // configuration - stack
 public:
  StackAutomata();
  ~StackAutomata();
  bool removeState(State state);
  // removes state state
  // returns: if successful
  // errors: E_A_STATE_BAD
  bool removeEntrySymbol(EntrySymbol symbol);
  // removes entry symbol symbol
  // returns: if successful
  // errors: E_A_ENTRY_SYMBOL_BAD
  StackSymbol addStackSymbol();
  // adds new stack symbol
  // returns: new stack symbol
  // errors: E_AS_STACK_SYMBOL_MAX
  bool removeStackSymbol(StackSymbol symbol);
  // removes stack symbol symbol
  // returns: if successful
  // errors: E_AS_STACK_SYMBOL_BAD
  set<StackSymbol> getStackSymbols();
  // returns: all stack symbols
  bool removeTransition(StackTransition transition);
  // removes transition transition
  // returns: if successful
  // errors: E_A_TRANSITION_BAD
  set<StackTransition> getTransitions();
  // returns: all transitions
  bool setStartStackSymbol(StackSymbol symbol);
  // sets start stack symbol to symbol
  // returns: if successful
  // errors: E_AS_STACK_SYMBOL_BAD
  StackSymbol getStartStackSymbol();
  // returns: start stack symbol
  bool addTransition(StackTransition &transition, State from_state,
		     EntrySymbol symbol, State to_state,
		     vector<StackSymbol> from_stack,
		     vector<StackSymbol> to_stack);
  // adds new transition consisting of from_state, symbol, to_state,
  // from_stack and to_stack
  // returns: if successful, in transition new transition
  // errors: E_A_STATE_BAD, E_A_ENTRY_SYMBOL_BAD, E_AS_STACK_SYMBOL_BAD
  bool addTransition(State from_state, EntrySymbol symbol, State to_state,
		     vector<StackSymbol> from_stack,
		     vector<StackSymbol> to_stack);
  // adds new transition consisting of from_state, symbol, to_state,
  // from_stack and to_stack
  // returns: if successful
  // errors: E_A_STATE_BAD, E_A_ENTRY_SYMBOL_BAD, E_AS_STACK_SYMBOL_BAD
  bool removeTransition(State from_state, EntrySymbol symbol, State to_state,
			vector<StackSymbol> from_stack,
			vector<StackSymbol> to_stack);
  // removes transition consisting of from_state, symbol, to_state,
  // from_stack and to_stack
  // returns: if successful
  // errors: E_A_STATE_BAD, E_A_ENTRY_SYMBOL_BAD, E_AS_STACK_SYMBOL_BAD,
  // E_A_TRANSITION_BAD
  bool setConfiguration(State state, vector<EntrySymbol> sentence,
			vector<StackSymbol> stack);
  // sets configuration consisting of state, sentence and stack
  // returns: if successful
  // errors: E_A_STATE_BAD, E_A_ENTRY_SYMBOL_BAD, E_AS_STACK_SYMBOL_BAD
  void getConfiguration(State& state, vector<EntrySymbol>& sentence,
			vector<StackSymbol>& stack);
  // returns: configuration state in state, sentence in sentence and stack in
  // stack
  set<vector<StackSymbol> > getFromStacks();
  // returns: stacks, which can be popped from stack (from all transitions)
  map<State, vector<StackSymbol> > getTransitionDestinations
    (vector<StackSymbol> from_stack);
  // returns: state and stack pairs, into which automata can do one transition
  // from actual configuration when popping from_stack from stack
  bool doTransition(State state, vector<StackSymbol> from_stack,
		    vector<StackSymbol> to_stack);
  // does one transition to state state from actual configuration when popping
  // from_stack and pushing to_stack to stack
  // returns: if successful
  // errors: E_AS_STATE_STACK_CANT_REACH, E_AS_STACK_CANT_POP
  bool checkSentence(vector<EntrySymbol> sentence);
  // returns: if sentence sentence is accepted by automata
  bool isDeterministic();
  // returns: if automata is deterministic
  bool isTopStackWorking();
  // returns: if automata is top stack working
  void toTopStackWorking(State &old_start_state,
			 StackSymbol &old_start_stack_symbol,
			 map<State, vector<State> > &new_states);
  // converts automata to equivalent top stack working one
  // returns: old start state in old_start_state, old start stack symbol in
  // old_start_stack_symbol and map of new state to vector of old states pairs
  bool isEmptyStackChecking();
  // returns: if automata is empty stack checking the sentence
  void toEmptyStackChecking(State &old_start_state,
			    StackSymbol &old_start_stack_symbol,
			    State &new_state);
  // converts automata to equivalent empty stack checking one
  // returns: old start state in old_start_state, old start stack symbol in
  // old_start_stack_symbol and new state in new_state
  void toEndStatesChecking(State &old_start_state,
			   StackSymbol &old_start_stack_symbol,
			   State &end_state);
  // converts automata to equivalent end states checking one
  // returns: old start state in old_start_state, old start stack symbol in
  // old_start_stack_symbol and new end state in end_state
 protected:
  set<StackTransition> find_transitions_with_state(State state);
  set<StackTransition> find_transitions_with_entry_symbol(EntrySymbol symbol);
  set<StackTransition> find_transitions_with_stack_symbol(StackSymbol symbol);
  set<vector<StackSymbol> > _getFromStacks(State state,
					   vector<StackSymbol> stack);
  map<State, vector<StackSymbol> > _getTransitionDestinations
    (State state, EntrySymbol symbol, vector<StackSymbol> stack);
  bool _doTransition(State& from_state, State to_state,
		     vector<EntrySymbol>& sentence,
		     vector<StackSymbol> from_stack,
		     vector<StackSymbol> to_stack, vector<StackSymbol> &stack);
  bool _checkSentence(State from_state, vector<EntrySymbol> sentence,
		      vector<StackSymbol> from_stack,
		      vector<StackSymbol>& stack);
  bool isPrefix12or21(vector<StackSymbol> stack1, vector<StackSymbol> stack2);
  void newStart(State &old_start_state, StackSymbol &old_start_stack_symbol);
};

#endif
