/* GAAE - GAAE Ain't an Automata Editor
 * Copyright (C) 2001 Jan Outrata <outrataj@upcase.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "error.h"
#include "stack_automata_l.h"

const char string_space = ' ';
const char string_lbracket = '(';
const char string_rbracket = ')';

StackAutomata_l::StackAutomata_l()
{
  automata = new StackAutomata;
  Automata_l::automata = automata;
  // setup EntrySymbol_NULL look
  entry_symbols[EntrySymbol_NULL] = EntrySymbol_l("-");
}

StackAutomata_l::~StackAutomata_l()
{
  delete automata;
  automata = NULL;
  Automata_l::automata = NULL;
}

string
StackAutomata_l::addStackSymbol(string name)
{
  string result;
  // check if we already don't have stack symbol with such name
  map<StackSymbol, StackSymbol_l>::iterator iter_map = stack_symbols.begin();
  for (; iter_map != stack_symbols.end(); ++iter_map)
    if (iter_map->second.getName() == name) break;
  if (iter_map != stack_symbols.end()) { 
    setError(E_AS_STACK_SYMBOL_NAME_EXISTS); return result;
  }
  StackSymbol new_symbol = automata->addStackSymbol();
  if (new_symbol == StackSymbol_NULL) return result;
  StackSymbol_l new_symbol_l(name);
  stack_symbols[new_symbol] = new_symbol_l;
  // return look
  result = new_symbol_l.getLook(new_symbol, getStackSymbolFlags(new_symbol));
  return result;
}

StackSymbol_l::t_flags
StackAutomata_l::getStackSymbolFlags(StackSymbol symbol)
{
  StackSymbol_l::t_flags result = (StackSymbol_l::t_flags)0;
  if (automata->getStartStackSymbol() == symbol) 
    result = (StackSymbol_l::t_flags)(result | StackSymbol_l::F_START);
  return result;
}

bool
StackAutomata_l::removeStackSymbol(string symbol_id)
{
  StackSymbol symbol = string_to_StackSymbol(symbol_id);
  if (automata->removeStackSymbol(symbol)) { 
    stack_symbols.erase(symbol); return true;
  }
  return false;
}

StackSymbol
StackAutomata_l::string_to_StackSymbol(string symbol_id)
{
  StackSymbol result = 0;
  int mult_factor = 1;
  reverse(symbol_id.begin(), symbol_id.end());
  for (string::const_iterator iter_s = symbol_id.begin();
       iter_s != symbol_id.end(); ++iter_s, mult_factor *= 10)
    result += (*iter_s - '0') * mult_factor;
  return result;
}

string
StackAutomata_l::editStackSymbol(string symbol_id, string new_name)
{
  string result;
  StackSymbol symbol = string_to_StackSymbol(symbol_id);
  map<StackSymbol, StackSymbol_l>::iterator 
    iter_sy_f = stack_symbols.find(symbol);
  if (iter_sy_f == stack_symbols.end()) { 
    setError(E_AS_STACK_SYMBOL_BAD); return result;
  }
  // check if we already don't have stack symbol with such name
  iter_sy_f = stack_symbols.begin();
  for (; iter_sy_f != stack_symbols.end(); ++iter_sy_f)
    if (iter_sy_f->second.getName() == new_name) break;
  if (iter_sy_f != stack_symbols.end()) { 
    setError(E_AS_STACK_SYMBOL_NAME_EXISTS); return result;
  }
  StackSymbol_l new_symbol_l(new_name);
  stack_symbols[symbol] = new_symbol_l;
  // return look
  result = new_symbol_l.getLook(symbol, getStackSymbolFlags(symbol));
  return result;
}

inline bool
StackAutomata_l::setStartStackSymbol(string symbol_id)
{
  return automata->setStartStackSymbol(string_to_StackSymbol(symbol_id));
}

bool
StackAutomata_l::addTransition(string from_state_id, string symbol_id,
			       string to_state_id, string from_stack,
			       string to_stack)
{
  vector<StackSymbol> from_stk, to_stk;
  string symbol;
  // parse stack strings
  for (string::const_iterator iter_s = from_stack.begin();
       iter_s != from_stack.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      from_stk.push_back(string_to_StackSymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) from_stk.push_back(string_to_StackSymbol(symbol));
  symbol.clear();
  for (string::const_iterator iter_s = to_stack.begin();
       iter_s != to_stack.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      to_stk.push_back(string_to_StackSymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) to_stk.push_back(string_to_StackSymbol(symbol));
  return automata->addTransition
    (string_to_State(from_state_id), string_to_EntrySymbol(symbol_id),
     string_to_State(to_state_id), from_stk, to_stk);
}

bool
StackAutomata_l::removeTransition(string from_state_id, string symbol_id,
				  string to_state_id, string from_stack,
				  string to_stack)
{
  vector<StackSymbol> from_stk, to_stk;
  string symbol;
  // parse stack strings
  for (string::const_iterator iter_s = from_stack.begin();
       iter_s != from_stack.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      from_stk.push_back(string_to_StackSymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) from_stk.push_back(string_to_StackSymbol(symbol));
  symbol.clear();
  for (string::const_iterator iter_s = to_stack.begin();
       iter_s != to_stack.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      to_stk.push_back(string_to_StackSymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) to_stk.push_back(string_to_StackSymbol(symbol));
  return automata->removeTransition
    (string_to_State(from_state_id), string_to_EntrySymbol(symbol_id),
     string_to_State(to_state_id), from_stk, to_stk);
}

bool
StackAutomata_l::setConfiguration(string state_id, string sentence,
				  string stack)
{
  vector<EntrySymbol> sent;
  string symbol;
  // parse sentence string
  for (string::const_iterator iter_s = sentence.begin();
       iter_s != sentence.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      sent.push_back(string_to_EntrySymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) sent.push_back(string_to_EntrySymbol(symbol));
  vector<StackSymbol> stk;
  // parse stack string
  symbol.clear();
  for (string::const_iterator iter_s = stack.begin(); iter_s != stack.end();
       ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      stk.push_back(string_to_StackSymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) stk.push_back(string_to_StackSymbol(symbol));
  return automata->setConfiguration(string_to_State(state_id), sent, stk);
}

void
StackAutomata_l::getConfiguration(string& state, string& sentence,
				  string& stack)
{
  State st;
  vector<EntrySymbol> sent;
  vector<StackSymbol> stk;
  automata->getConfiguration(st, sent, stk);
  // return st, sent and stk looks
  state = states[st].getLook(st, getStateFlags(st));
  sentence = string_lbracket; sentence += "entry_symbols";
  for (vector<EntrySymbol>::const_iterator iter_sy = sent.begin();
       iter_sy != sent.end(); ++iter_sy)
    sentence += string_space + entry_symbols[*iter_sy].getLook(*iter_sy);
  sentence += string_rbracket;
  stack = string_lbracket; stack += "stack";
  for (vector<StackSymbol>::const_iterator iter_sy = stk.begin();
       iter_sy != stk.end(); ++iter_sy)
    stack += string_space + 
      stack_symbols[*iter_sy].getLook(*iter_sy, getStackSymbolFlags(*iter_sy));
  stack += string_rbracket;
}

string
StackAutomata_l::getLook()
{
  string result;
  result += string_lbracket; result += "stack_automata";
  result += string_space;
  // State looks
  result += string_lbracket; result += "states";
  set<State> sts = automata->getStates();
  for (set<State>::const_iterator iter_st = sts.begin(); iter_st != sts.end();
       ++iter_st)
    result += string_space +
      states[*iter_st].getLook(*iter_st, getStateFlags(*iter_st));
  result += string_rbracket; result += string_space;
  // EntrySymbol looks
  result += string_lbracket; result += "entry_symbols";
  set<EntrySymbol> e_symbols = automata->getEntrySymbols();
  for (set<EntrySymbol>::const_iterator iter_sy = e_symbols.begin();
       iter_sy != e_symbols.end(); ++iter_sy)
    result += string_space + entry_symbols[*iter_sy].getLook(*iter_sy);
  result += string_rbracket; result += string_space;
  // StackSymbol looks
  result += string_lbracket; result += "stack_symbols";
  set<StackSymbol> s_symbols = automata->getStackSymbols();
  for (set<StackSymbol>::const_iterator iter_sy = s_symbols.begin();
       iter_sy != s_symbols.end(); ++iter_sy)
    result += string_space + 
      stack_symbols[*iter_sy].getLook(*iter_sy, getStackSymbolFlags(*iter_sy));
  result += string_rbracket; result += string_space;
  // transition looks
  // transition look: (transition 
  // _from_state_look_ _entry_symbol_look_ _to_state_look_ 
  // (from_stack _from_stack_symbol_look_ _from_stack_symbol_look_) 
  // (to_stack _to_stack_symbol_look_ _to_stack_symbol_look_)) 
  result += string_lbracket; result += "transitions";
  set<StackTransition> trans = automata->getTransitions();
  for (set<StackTransition>::const_iterator iter_t = trans.begin();
       iter_t != trans.end(); ++iter_t) {
    result += string_space; result += string_lbracket; result += "transition";
    result += string_space + states[iter_t->from_state].getLook
      (iter_t->from_state, getStateFlags(iter_t->from_state));
    result += string_space + entry_symbols[iter_t->entry_symbol].getLook
      (iter_t->entry_symbol);
    result += string_space + states[iter_t->to_state].getLook
      (iter_t->to_state, getStateFlags(iter_t->to_state));
    result += string_space; result += string_lbracket; result += "from_stack";
    for (vector<StackSymbol>::const_iterator
	   iter_sy = iter_t->from_stack.begin(); 
	 iter_sy != iter_t->from_stack.end(); ++iter_sy)
      result += string_space + 
	stack_symbols[*iter_sy].getLook
	(*iter_sy, getStackSymbolFlags(*iter_sy));
    result += string_rbracket;
    result += string_space; result += string_lbracket; result += "to_stack";
    for (vector<StackSymbol>::const_iterator
	   iter_sy = iter_t->to_stack.begin();
	 iter_sy != iter_t->to_stack.end(); ++iter_sy)
      result += string_space + 
	stack_symbols[*iter_sy].getLook
	(*iter_sy, getStackSymbolFlags(*iter_sy));
    result += string_rbracket;
    result += string_rbracket;
  }
  result += string_rbracket; result += string_rbracket;
  return result;
}

string
StackAutomata_l::getFromStacks()
{
  string result;
  result += string_lbracket; result += "stacks"; result += string_space; 
  set<vector<StackSymbol> > stacks = automata->getFromStacks();
  for (set<vector<StackSymbol> >::const_iterator iter_v = stacks.begin();
       iter_v != stacks.end(); ++iter_v) {
    result += string_lbracket; result += "stack";
    for (vector<StackSymbol>::const_iterator iter_sy = iter_v->begin(); 
	 iter_sy != iter_v->end(); ++iter_sy)
      result += string_space + 
	stack_symbols[*iter_sy].getLook
	(*iter_sy, getStackSymbolFlags(*iter_sy));
    result += string_rbracket; result += string_space;
  }
  string::iterator last = result.end(); --last;
  result.erase(last); result += string_rbracket;
  return result;
}

string
StackAutomata_l::getTransitionDestinations(string from_stack)
{
  string result;
  vector<StackSymbol> from_stk;
  string symbol;
  // parse stack string
  for (string::const_iterator iter_s = from_stack.begin();
       iter_s != from_stack.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      from_stk.push_back(string_to_StackSymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) from_stk.push_back(string_to_StackSymbol(symbol));
  map<State, vector<StackSymbol> > 
    dests = automata->getTransitionDestinations(from_stk);
  // return dests look
  result += string_lbracket; result += "state-stack-pairs";
  result += string_space; 
  State st;
  vector<StackSymbol> stk;
  for (map<State, vector<StackSymbol> >::const_iterator 
	 iter_map = dests.begin(); iter_map != dests.end(); ++iter_map) {
    result += string_lbracket;
    st = iter_map->first;
    result += states[st].getLook(st, getStateFlags(st)) + string_space;
    stk = iter_map->second;
    result += string_lbracket; result += "stack";
    for (vector<StackSymbol>::const_iterator iter_sy = stk.begin();
	 iter_sy != stk.end(); ++iter_sy)
      result += string_space + 
	stack_symbols[*iter_sy].getLook
	(*iter_sy, getStackSymbolFlags(*iter_sy));
    result += string_rbracket;
    result += string_rbracket; result += string_space;
  }
  string::iterator last = result.end(); --last;
  result.erase(last); result += string_rbracket;
  return result;
}

bool
StackAutomata_l::doTransition(string state_id, string from_stack,
			      string to_stack)
{
  vector<StackSymbol> from_stk, to_stk;
  string symbol;
  // parse stack strings
  for (string::const_iterator iter_s = from_stack.begin();
       iter_s != from_stack.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      from_stk.push_back(string_to_StackSymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) from_stk.push_back(string_to_StackSymbol(symbol));
  symbol.clear();
  for (string::const_iterator iter_s = to_stack.begin();
       iter_s != to_stack.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      to_stk.push_back(string_to_StackSymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) to_stk.push_back(string_to_StackSymbol(symbol));
  return automata->doTransition(string_to_State(state_id), from_stk, to_stk);
}

bool
StackAutomata_l::checkSentence(string sentence)
{
  vector<EntrySymbol> sent;
  string symbol;
  // parse sentence string
  for (string::const_iterator iter_s = sentence.begin();
       iter_s != sentence.end(); ++iter_s)
    if (*iter_s != string_space) symbol += *iter_s;
    else {
      sent.push_back(string_to_EntrySymbol(symbol));
      symbol.clear();
    }
  if (!symbol.empty()) sent.push_back(string_to_EntrySymbol(symbol));
  return automata->checkSentence(sent);
}

inline bool
StackAutomata_l::isDeterministic()
{
  return automata->isDeterministic();
}

inline bool
StackAutomata_l::isTopStackWorking()
{
  return automata->isTopStackWorking();
}

void
StackAutomata_l::toTopStackWorking()
{
  if (isTopStackWorking()) return;
  State old_start_state;
  StackSymbol old_start_stack_symbol;
  map<State, vector<State> > new_states;
  automata->toTopStackWorking
    (old_start_state, old_start_stack_symbol, new_states);
  string new_name = states[old_start_state].getName() + "'";
  states[automata->getStartState()] = State_l(new_name);
  new_name = stack_symbols[old_start_stack_symbol].getName() + "'";
  stack_symbols[automata->getStartStackSymbol()] = StackSymbol_l(new_name);
  for (map<State, vector<State> >::const_iterator
	 iter_map = new_states.begin(); iter_map != new_states.end();
       ++iter_map) {
    new_name = states[iter_map->first].getName();
    int i = 1;
    for (vector<State>::const_iterator iter_st = iter_map->second.begin();
	 iter_st != iter_map->second.end(); ++iter_st, ++i)
      states[*iter_st] = State_l(new_name + "_" + int_to_string(i));
  }
}

string
StackAutomata_l::int_to_string(int i)
{
  string result;
  for (; i > 0; i /= 10) result += i % 10 + '0';
  if (result.empty()) result = '0';
  reverse(result.begin(), result.end());
  return result;
}

inline bool
StackAutomata_l::isEmptyStackChecking()
{
  return automata->isEmptyStackChecking();
}

void
StackAutomata_l::toEmptyStackChecking()
{
  if (isEmptyStackChecking()) return;
  State old_start_state, new_state;
  StackSymbol old_start_stack_symbol;
  automata->toEmptyStackChecking
    (old_start_state, old_start_stack_symbol, new_state);
  string new_name = states[old_start_state].getName() + "'";
  states[automata->getStartState()] = State_l(new_name);
  new_name = stack_symbols[old_start_stack_symbol].getName() + "'";
  stack_symbols[automata->getStartStackSymbol()] = StackSymbol_l(new_name);
  new_name = "F";
  states[new_state] = State_l(new_name);
}

void
StackAutomata_l::toEndStatesChecking()
{
  if (!isEmptyStackChecking()) return;
  State old_start_state, end_state;
  StackSymbol old_start_stack_symbol;
  automata->toEndStatesChecking
    (old_start_state, old_start_stack_symbol, end_state);
  string new_name = states[old_start_state].getName() + "'";
  states[automata->getStartState()] = State_l(new_name);
  new_name = stack_symbols[old_start_stack_symbol].getName() + "'";
  stack_symbols[automata->getStartStackSymbol()] = StackSymbol_l(new_name);
  new_name = "F";
  states[end_state] = State_l(new_name);
}

const StackAutomata_l &
StackAutomata_l::operator=(const StackAutomata_l &a)
{
  if (this != &a) {
    states = a.states;
    entry_symbols = a.entry_symbols;
    stack_symbols = a.stack_symbols;
    *automata = *(a.automata);
    *(Automata_l::automata) = *(a.automata);
  }
  return *this;
}
